import re
import sys
import os
import datetime
import tkinter as tk
from tkinter import ttk, messagebox, filedialog

from mmg_toolbox.utils.env_functions import run_python_string


# Define colors for the various types of tokens
class Colours:
    normal = '#eaeaea'  # rgb((234, 234, 234))
    keywords = '#ea5f5f'  # rgb((234, 95, 95))
    commands = '#54AAE3'
    comments = '#5feaa5'  # rgb((95, 234, 165))
    string = '#eaa25f'  # rgb((234, 162, 95))
    function = '#5fd3ea'  # rgb((95, 211, 234))
    background = '#2a2a2a'  # rgb((42, 42, 42))

FONT = 'Consolas 15'
TAB_WIDTH = 4
INDENT = ' ' * TAB_WIDTH

# Define a list of Regex Pattern that should be colored in a certain way
REPL = [
    [
        r'(?:^|\s|\W)(False|None|True|and|as|assert|async|await|' +
        'break|class|continue|def|del|elif|else|except|finally|for|from|' +
        'global|if|import|in|is|lambda|nonlocal|not|or|pass|print|' +
        r'raise|return|try|while|with|yield)(?:$|\s|\W)',
        Colours.keywords
    ],
    # [r'(?:^|\s|\W)(scan|scancn|cscan|frange|pos|inc|go)(?:$|\s|\W)', commands],
    ['".*?"', Colours.string],
    ['\'.*?\'', Colours.string],
    ['#.*?$', Colours.comments],
]

SCRIPT = '''"""
Example Script
%s
"""

pos shutter 1
pos x1 1

scancn eta 0.01 101 pil 1 roi2

for chi_val in frange(84, 96, 2):
    pos chi chi_val
    scancn eta 0.01 101 pil 1 roi2
'''


def search_re(pattern, text):
    matches = []
    text = text.splitlines()
    regex = re.compile(pattern)
    for i, line in enumerate(text):
        for match in regex.finditer(line):
            if match.groups():
                matches.append((f"{i + 1}.{match.span(1)[0]}", f"{i + 1}.{match.span(1)[1]}"))
            else:
                matches.append((f"{i + 1}.{match.start()}", f"{i + 1}.{match.end()}"))
    return matches


def default_script():
    return SCRIPT % datetime.datetime.now().strftime('%Y-%m-%d %H:%M')


class PythonEditorFrame:
    """
    Editable textbox with numbers at side and key bindings for Python
    """

    def __init__(self, root: tk.Misc | tk.Tk, script_string=None, config: dict | None = None,
                 filename: str | None = None):
        self.root = root
        self.config = config
        # Variables
        self.filename = filename or ''
        self.script_string = script_string or default_script()

        "----------- Textbox -----------"

        txt = ttk.Frame(root)
        txt.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)

        # Scrollbars
        scanx = ttk.Scrollbar(txt, orient=tk.HORIZONTAL)
        scanx.pack(side=tk.BOTTOM, fill=tk.X)
        scany = ttk.Scrollbar(txt, orient=tk.VERTICAL)
        scany.pack(side=tk.RIGHT, fill=tk.Y)

        # Text numbers
        border = 10
        self.textno = tk.Text(txt, width=3, font=FONT, borderwidth=border, relief=tk.FLAT)
        self.textno.pack(side=tk.LEFT, fill=tk.Y, expand=tk.YES)
        self.textno.config(yscrollcommand=scany.set, state=tk.DISABLED)

        # TEXT box
        # Add a hefty border width so we can achieve a little bit of padding
        self.text = tk.Text(
            txt,
            background=Colours.background,
            foreground=Colours.normal,
            insertbackground=Colours.normal,
            relief=tk.FLAT,
            borderwidth=border,
            font=FONT,
            undo=True,
            autoseparators=True,
            maxundo=-1,
            wrap=tk.NONE
        )
        self.text.pack(side=tk.LEFT, fill=tk.BOTH, expand=tk.YES)
        self.text.insert('1.0', self.script_string)
        self.text.bind('<KeyRelease>', self.changes)
        self.text.bind('<Return>', self.auto_indent)
        self.text.bind('<KP_Enter>', self.auto_indent)
        self.text.bind('<Tab>', self.tab)
        self.text.bind('<Shift-Tab>', self.shift_tab)
        self.text.bind('<Control-slash>', self.comment)
        self.text.bind('<BackSpace>', self.delete)

        # make scrollbars work
        self.text.config(xscrollcommand=scanx.set, yscrollcommand=scany.set)
        scanx.config(command=self.text.xview)
        scany.config(command=self.text.yview)

        frm = ttk.Frame(root, relief=tk.RIDGE, borderwidth=2)
        frm.pack(side=tk.TOP, fill=tk.BOTH, expand=tk.YES)
        ttk.Button(frm, text='RUN', command=self.run).pack(side=tk.RIGHT, pady=5)


        self.changes()

    "------------------------------------------------------------------------"
    "-------------------------Load/Save Functions----------------------------"
    "------------------------------------------------------------------------"

    def set_filename(self, filename: str):
        self.filename = filename
        if hasattr(self.root, 'wm_title'):
            self.root.wm_title(filename)

    def new(self):
        answer = messagebox.askokcancel(
            title='Script editor',
            message='Do you want to replace the current script?',
            parent=self.root,
        )
        if answer:
            self.set_filename('new_file.py')
            self.text.delete('1.0', tk.END)
            self.text.insert('1.0', default_script())
            self.changes()

    def open(self):
        """Open new script"""
        filename = filedialog.askopenfilename(
            title='Open Python Script',
            initialdir=os.path.dirname(self.filename),
            defaultextension='*.py',
            filetypes=(("Python files", "*.py"), ("All files", "*.*"))
        )
        if filename:
            with open(filename, 'r') as f:
                self.script_string = f.read()
            self.text.delete('1.0', tk.END)
            self.text.insert('1.0', self.script_string)
            self.set_filename(filename)
            self.changes()

    def saveas(self):
        """Save as file"""
        filename = filedialog.asksaveasfilename(
            title='Python Script',
            initialfile=self.filename,
            defaultextension='.py'
        )
        if filename:
            with open(filename, 'w') as f:
                f.write(self.script_string)
            print('Written script to %s' % filename)
            self.set_filename(filename)

    def save(self):
        """Save script"""
        if self.filename == '':
            self.saveas()
            return
        with open(self.filename, 'w') as f:
            f.write(self.script_string)
        print('Written script to %s' % self.filename)

    def run(self):
        """Run script"""
        run_python_string(self.script_string)

    "------------------------------------------------------------------------"
    "--------------------------General Functions-----------------------------"
    "------------------------------------------------------------------------"

    def changes(self, event=None):
        """ Register Changes made to the Editor Content """

        # If actually no changes have been made stop / return the function
        if self.text.get('1.0', tk.END) == self.script_string:
            return

        # Remove all tags so they can be redrawn
        for tag in self.text.tag_names():
            self.text.tag_remove(tag, "1.0", tk.END)

        # Add tags where the search_re function found the pattern
        i = 0
        for pattern, color in REPL:
            for start, end in search_re(pattern, self.text.get('1.0', tk.END)):
                self.text.tag_add(f'{i}', start, end)
                self.text.tag_config(f'{i}', foreground=color)
                i += 1

        # Add tags to multiline comments
        start = None
        for n, line in enumerate(self.text.get('1.0', tk.END).splitlines()):
            for match in re.finditer('\'{3}|\"{3}', line):
                if start:
                    self.text.tag_add(f'{i}', start, f"{n + 1}.{match.end()}")
                    self.text.tag_config(f'{i}', foreground=Colours.comments)
                    i += 1
                    start = None
                else:
                    start = f"{n + 1}.{match.start()}"

        self.script_string = self.text.get('1.0', tk.END)

        self.textno.configure(state='normal')
        self.textno.replace('1.0', tk.END, '\n'.join(str(n+1) for n in range(self.script_string.count('\n'))))
        self.textno.configure(state='disabled')

    def tab(self, event=None):
        if event is None:
            text = self.text
        else:
            text = event.widget
        try:
            first = int(text.index(tk.SEL_FIRST).split('.')[0])
            last = int(text.index(tk.SEL_LAST).split('.')[0])
            for lineno in range(first, last + 1):
                text.insert('%d.0' % lineno, INDENT)
        except tk.TclError:
            text.insert(tk.INSERT, INDENT)
        return 'break'

    def shift_tab(self, event=None):
        if event is None:
            text = self.text
        else:
            text = event.widget
        try:  # selection
            first = int(text.index(tk.SEL_FIRST).split('.')[0])
            last = int(text.index(tk.SEL_LAST).split('.')[0])
            for lineno in range(first, last + 1):
                line = text.get('%d.0' % lineno, '%d.0 lineend' % lineno)
                spaceno = len(line) - len(line.lstrip())
                spaceno = 4 if spaceno > 4 else spaceno
                text.delete('%d.0' % lineno, '%d.%d' % (lineno, spaceno))
        except tk.TclError:  # single point
            lineno = int(text.index('insert').split('.')[0])
            line = text.get('%d.0' % lineno, '%d.0 lineend' % lineno)
            spaceno = len(line) - len(line.lstrip())
            text.delete('%d.0' % lineno, '%d.%d' % (lineno, spaceno))
        return 'break'

    def comment(self, event=None):
        if event is None:
            text = self.text
        else:
            text = event.widget
        try:
            first = int(text.index(tk.SEL_FIRST).split('.')[0])
            last = int(text.index(tk.SEL_LAST).split('.')[0])
            line = text.get('%d.0' % first, '%d.0 lineend' % first)
            if line.startswith('#'):  # remove comments
                for lineno in range(first, last + 1):
                    line = text.get('%d.0' % lineno, '%d.0 lineend' % lineno)
                    if line.startswith('# '):
                        text.delete('%d.0' % lineno, '%d.2' % lineno)
                    elif line.startswith('#'):
                        text.delete('%d.0' % lineno)
            else:  # add comments
                for lineno in range(first, last + 1):
                    line = text.get('%d.0' % lineno, '%d.0 lineend' % lineno)
                    if not line.startswith('#'):
                        text.insert('%d.0' % lineno, '# ')
        except tk.TclError:
            lineno = int(text.index('insert').split('.')[0])
            line = text.get('%d.0' % lineno, '%d.0 lineend' % lineno)
            if line.startswith('# '):
                text.delete('%d.0' % lineno, '%d.2' % lineno)
            elif line.startswith('#'):
                text.delete('%d.0' % lineno)
            else:
                text.insert('%d.0' % lineno, '# ')
        return 'break'

    def auto_indent(self, event=None):
        if event is None:
            text = self.text
        else:
            text = event.widget

        # get leading whitespace from current line
        line = text.get("insert linestart", "insert")
        match = re.match(r'^(\s+)', line)
        whitespace = match.group(0) if match else ""
        if any(line.strip().endswith(c) for c in [':', '{', '[', '(']):
            whitespace += INDENT

        # insert the newline and the whitespace
        text.insert("insert", f"\n{whitespace}")

        # return "break" to inhibit default insertion of newline
        return "break"

    def delete(self, event=None):
        if event is None:
            text = self.text
        else:
            text = event.widget

        try:
            text.delete(tk.SEL_FIRST, tk.SEL_LAST)
        except tk.TclError:
            line = text.get("insert linestart", "insert")
            previous = text.get("insert -%d chars" % TAB_WIDTH, "insert")
            if line == " " * len(line) and len(line) % TAB_WIDTH > 0:  # delete tab
                text.delete("insert -%d chars" % (len(line) % TAB_WIDTH), "insert")
            elif previous == " " * TAB_WIDTH:  # delete tab
                text.delete("insert-%d chars" % TAB_WIDTH, "insert")
            elif '\n' in previous and previous[-1] != '\n':  # delete spaces to start of line
                text.delete("insert-%d chars" % len(line), "insert")
            else:  # normal delete
                text.delete("insert-1 chars", "insert")
        return "break"


class PythonTerminalFrame:
    """
    Editable textbox with numbers at side and key bindings for Python
    UNFINISHED
    """

    def __init__(self, root: tk.Misc):

        txt = ttk.Frame(root)
        txt.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)

        # Scrollbars
        scanx = ttk.Scrollbar(txt, orient=tk.HORIZONTAL)
        scanx.pack(side=tk.BOTTOM, fill=tk.X)
        scany = ttk.Scrollbar(txt, orient=tk.VERTICAL)
        scany.pack(side=tk.RIGHT, fill=tk.Y)

        # Text numbers
        border = 10
        self.textno = tk.Text(txt, width=3, font=FONT, borderwidth=border, relief=tk.FLAT)
        self.textno.pack(side=tk.LEFT, fill=tk.Y, expand=tk.NO)
        self.textno.config(yscrollcommand=scany.set, state=tk.DISABLED)

        # Terminal
        frm = ttk.Frame(txt)
        frm.pack(side=tk.LEFT, fill=tk.BOTH, expand=tk.YES)
        # history
        self.history_str = f"Python {sys.version} on {sys.platform}\n"
        self.text = tk.Text(
            frm,
            background=Colours.background,
            foreground=Colours.normal,
            insertbackground=Colours.normal,
            relief=tk.FLAT,
            borderwidth=border,
            font=FONT,
            undo=True,
            autoseparators=True,
            maxundo=-1,
            wrap=tk.NONE
        )
        self.text.pack(side=tk.LEFT, fill=tk.BOTH, expand=tk.YES)
        self.text.insert('1.0', self.history_str)
        self.text.config(xscrollcommand=scanx.set, yscrollcommand=scany.set, state=tk.DISABLED)
        # entry
        self.entry = tk.Text(
            frm,
            background=Colours.background,
            foreground=Colours.normal,
            insertbackground=Colours.normal,
            relief=tk.FLAT,
            borderwidth=border,
            font=FONT,
            undo=True,
            autoseparators=True,
            maxundo=-1,
            wrap=tk.NONE,
            height=2
        )
        self.entry.pack(side=tk.LEFT, fill=tk.BOTH, expand=tk.YES)
        self.entry.insert('1.0', ">>> ")
        self.entry.bind('<KeyRelease>', self.changes)
        self.entry.bind('<Return>', self.auto_indent)
        self.entry.bind('<KP_Enter>', self.auto_indent)
        self.entry.bind('<Tab>', self.tab)
        self.entry.bind('<Shift-Tab>', self.shift_tab)
        self.entry.bind('<Control-slash>', self.comment)
        self.entry.bind('<BackSpace>', self.delete)

        # make scrollbars work
        self.text.config(xscrollcommand=scanx.set, yscrollcommand=scany.set)
        scanx.config(command=self.text.xview)
        scany.config(command=self.text.yview)

        self.changes()

    "------------------------------------------------------------------------"
    "--------------------------General Functions-----------------------------"
    "------------------------------------------------------------------------"

    def changes(self, event=None):
        """ Register Changes made to the Editor Content """

        # If actually no changes have been made stop / return the function
        if self.text.get('1.0', tk.END) == self.history_str:
            return

        # Remove all tags so they can be redrawn
        for tag in self.text.tag_names():
            self.text.tag_remove(tag, "1.0", tk.END)

        # Add tags where the search_re function found the pattern
        i = 0
        for pattern, color in REPL:
            for start, end in search_re(pattern, self.text.get('1.0', tk.END)):
                self.text.tag_add(f'{i}', start, end)
                self.text.tag_config(f'{i}', foreground=color)
                i += 1

        # Add tags to multiline comments
        start = None
        for n, line in enumerate(self.text.get('1.0', tk.END).splitlines()):
            for match in re.finditer('\'{3}|\"{3}', line):
                if start:
                    self.text.tag_add(f'{i}', start, f"{n + 1}.{match.end()}")
                    self.text.tag_config(f'{i}', foreground=Colours.comments)
                    i += 1
                    start = None
                else:
                    start = f"{n + 1}.{match.start()}"

        self.history_str = self.text.get('1.0', tk.END)

        self.textno.configure(state='normal')
        self.textno.replace('1.0', tk.END, '\n'.join(str(n + 1) for n in range(self.history_str.count('\n'))))
        self.textno.configure(state='disabled')

    def tab(self, event=None):
        if event is None:
            text = self.text
        else:
            text = event.widget
        try:
            first = int(text.index(tk.SEL_FIRST).split('.')[0])
            last = int(text.index(tk.SEL_LAST).split('.')[0])
            for lineno in range(first, last + 1):
                text.insert('%d.0' % lineno, INDENT)
        except tk.TclError:
            text.insert(tk.INSERT, INDENT)
        return 'break'

    def shift_tab(self, event=None):
        if event is None:
            text = self.text
        else:
            text = event.widget
        try:  # selection
            first = int(text.index(tk.SEL_FIRST).split('.')[0])
            last = int(text.index(tk.SEL_LAST).split('.')[0])
            for lineno in range(first, last + 1):
                line = text.get('%d.0' % lineno, '%d.0 lineend' % lineno)
                spaceno = len(line) - len(line.lstrip())
                spaceno = 4 if spaceno > 4 else spaceno
                text.delete('%d.0' % lineno, '%d.%d' % (lineno, spaceno))
        except tk.TclError:  # single point
            lineno = int(text.index('insert').split('.')[0])
            line = text.get('%d.0' % lineno, '%d.0 lineend' % lineno)
            spaceno = len(line) - len(line.lstrip())
            text.delete('%d.0' % lineno, '%d.%d' % (lineno, spaceno))
        return 'break'

    def comment(self, event=None):
        if event is None:
            text = self.text
        else:
            text = event.widget
        try:
            first = int(text.index(tk.SEL_FIRST).split('.')[0])
            last = int(text.index(tk.SEL_LAST).split('.')[0])
            line = text.get('%d.0' % first, '%d.0 lineend' % first)
            if line.startswith('#'):  # remove comments
                for lineno in range(first, last + 1):
                    line = text.get('%d.0' % lineno, '%d.0 lineend' % lineno)
                    if line.startswith('# '):
                        text.delete('%d.0' % lineno, '%d.2' % lineno)
                    elif line.startswith('#'):
                        text.delete('%d.0' % lineno)
            else:  # add comments
                for lineno in range(first, last + 1):
                    line = text.get('%d.0' % lineno, '%d.0 lineend' % lineno)
                    if not line.startswith('#'):
                        text.insert('%d.0' % lineno, '# ')
        except tk.TclError:
            lineno = int(text.index('insert').split('.')[0])
            line = text.get('%d.0' % lineno, '%d.0 lineend' % lineno)
            if line.startswith('# '):
                text.delete('%d.0' % lineno, '%d.2' % lineno)
            elif line.startswith('#'):
                text.delete('%d.0' % lineno)
            else:
                text.insert('%d.0' % lineno, '# ')
        return 'break'

    def auto_indent(self, event=None):
        if event is None:
            text = self.text
        else:
            text = event.widget

        # get leading whitespace from current line
        line = text.get("insert linestart", "insert")
        match = re.match(r'^(\s+)', line)
        whitespace = match.group(0) if match else ""
        if any(line.strip().endswith(c) for c in [':', '{', '[', '(']):
            whitespace += INDENT

        # insert the newline and the whitespace
        text.insert("insert", f"\n{whitespace}")

        # return "break" to inhibit default insertion of newline
        return "break"

    def delete(self, event=None):
        if event is None:
            text = self.entry
        else:
            text = event.widget

        try:
            text.delete(tk.SEL_FIRST, tk.SEL_LAST)
        except tk.TclError:
            line = text.get("insert linestart", "insert")
            previous = text.get("insert -%d chars" % TAB_WIDTH, "insert")
            if line == " " * len(line) and len(line) % TAB_WIDTH > 0:  # delete tab
                text.delete("insert -%d chars" % (len(line) % TAB_WIDTH), "insert")
            elif previous == " " * TAB_WIDTH:  # delete tab
                text.delete("insert-%d chars" % TAB_WIDTH, "insert")
            elif '\n' in previous and previous[-1] != '\n':  # delete spaces to start of line
                text.delete("insert-%d chars" % len(line), "insert")
            else:  # normal delete
                text.delete("insert-1 chars", "insert")
        return "break"

    def execute(self, event=None):
        if event is None:
            text = self.entry
        else:
            text = event.widget

        cmd = text.get('1.0', tk.END)
        cmd = cmd.strip('> ')  # remove preceding chevrons
        exec(cmd)



"""
msmapper utils
"""
import numpy as np
import h5py

from mmg_toolbox import version_info
from mmg_toolbox.nexus import nexus_writer as nw
from mmg_toolbox.utils.fitting import FitResults


MSMAPPER_VERSION = '1.9'


def create_bean(input_files, output_file, start=None, shape=None, step=None,
                output_mode=None, to_crystal=None, normalisation=None, polarisation=None,
                detector_region=None, reduce_box=None, third_axis=None,
                azi_plane_normal=None):
    """
    Create a bean file for msmapper in a temporary directory
     currently only allows a few standard inputs: hkl_start, shape and step values.
    :param input_files: list of scan file locations
    :param output_file: str location of output file
    :param start: [h, k, l] start of box (None to omit and calculate autobox)
    :param shape: [n, m, o] size of box in voxels (None to omit and calcualte autobox)
    :param step: [dh, dk, dl] step size in each direction - size of voxel in reciprocal lattice units
    :param output_mode: 'Volume_HKL' or 'Volume_Q' type of calculation
    :param to_crystal: for Volume_Q, use the crystal frame if True, or Lab frame otherwise
    :param normalisation: Monitor value to use for normalisation, e.g. 'rc'
    :param polarisation: Bool apply polarisation correction
    :param detector_region: [sx, ex, sy, ey] region of interest on detector
    :param reduce_box: Bool, reduce box to non-zero elements
    :param third_axis: [h, k, l] direction of Z-axis of voxel grid
    :param azi_plane_normal: [h, k, l] sets X-axis of voxel grid, normal to Z-axis
    :return: str file location of bean file
    """
    input_files = np.asarray(input_files, dtype=str).reshape(-1).tolist()
    # Remove empty entries
    while '' in input_files:
        input_files.remove('')

    if step is None:
        step = [0.001, 0.001, 0.001]
    else:
        step = np.asarray(step, dtype=float).reshape(-1).tolist()

    if shape is not None:
        shape = np.asarray(shape, dtype=int).reshape(-1).tolist()

    if start is not None:
        start = np.asarray(start, dtype=float).reshape(-1).tolist()

    if reduce_box is None:
        reduce_box = False

    bean = {
        "inputs": input_files,  # Filename of scan file
        "output": output_file,
        # Output filename - must be in processing directory, or somewhere you can write to
        "splitterName": "gaussian",  # one of the following strings "nearest", "gaussian", "negexp", "inverse"
        "splitterParameter": 2.0,
        # splitter's parameter is distance to half-height of the weight function.
        # If you use None or "" then it is treated as "nearest"
        "scaleFactor": 2.0,
        # the oversampling factor for each image; to ensure that are no gaps in between pixels in mapping
        "step": step,
        # a single value or list if 3 values and determines the lengths of each side of the voxels in the volume
        "start": start,  # location in HKL space of the bottom corner of the array.
        "shape": shape,  # size of the array to create for reciprocal space volume
        "reduceToNonZero": reduce_box  # True/False, if True, attempts to reduce the volume output
    }
    if output_mode:
        bean['outputMode'] = output_mode
    if MSMAPPER_VERSION > '1.8' and to_crystal is not None:
        bean['toCrystalFrame'] = to_crystal
    if normalisation:
        bean['monitorName'] = normalisation
    if polarisation:
        bean['correctPolarization'] = polarisation
    if detector_region:
        bean['region'] = detector_region
    if MSMAPPER_VERSION > '1.7' and third_axis:
        bean['thirdAxis'] = np.array(third_axis).tolist()
        bean['aziPlaneNormal'] = np.array(azi_plane_normal).tolist()
    return bean


def update_msmapper_nexus(filename: str, hkl_slice: tuple[np.ndarray, np.ndarray, np.ndarray],
                          orthogonal_axes: tuple[np.ndarray, np.ndarray, np.ndarray] | None = None,
                          average_axes: tuple[np.ndarray, np.ndarray, np.ndarray] | None = None,
                          fit_options: dict | None = None, h_result: FitResults | None = None,
                          k_result: FitResults | None = None, l_result: FitResults | None = None,
                          q_result: FitResults | None = None, tth_result: FitResults | None = None):

    """
    Update the NeXus file generated by msmapper with additional analysis data.

    This function opens an existing NeXus file produced by **msmapper** and appends
    analysis-related datasets and metadata. Typical additions include HKL slices,
    optional orthogonal/averaged axis data, fitting results (e.g., peak fits along
    H, K, L, Q, and 2θ), and any options/configuration used during fitting.

    The function is non-destructive with respect to the original raw datasets
    written by msmapper; new fields are appended under appropriate NXgroups.

    :param filename: Path to the NeXus (.nxs/.h5) file to be updated in-place.
    :type filename: str

    :param hkl_slice: slices of the reciprocal space volume in each of h, k, l directions
    :type hkl_slice: tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray]

    :param orthogonal_axes: Optional tuple of three arrays representing an
        orthogonal coordinate system (e.g., lab-frame axes) mapped to the same points
        as the HKL slice. If provided, these are stored alongside HKL for reference.
        Order is expected to be (axis_x, axis_y, axis_z) and shape must match volume.
    :type orthogonal_axes: tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray] | None

    :param average_axes: Optional tuple of three arrays representing averaged axes
        (e.g., binned or smoothed directions) corresponding to the HKL slice. Order
        is (wavevector==|Q|, two-theta [Deg], averaged intensities).
    :type average_axes: tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray] | None

    :param fit_options: Optional dictionary of fitting options used to produce the
        FitResults (e.g., model type, bounds, initial guesses, weighting). This is
        serialized into the NeXus file for provenance.
    :type fit_options: dict | None

    :param h_result: Peak Fit results for the H direction (e.g., peak positions, widths,
        amplitudes, residuals). If provided, these are written under the appropriate
        group in the NeXus file.
    :type h_result: FitResults | None

    :param k_result: Fit results for the K direction.
    :type k_result: FitResults | None

    :param l_result: Fit results for the L direction.
    :type l_result: FitResults | None

    :param q_result: Fit results for Q (|Q| or wavevector magnitude).
    :type q_result: FitResults | None

    :param tth_result: Fit results for two-theta (2θ).
    :type tth_result: FitResults | None

    :returns: None. The function updates the specified NeXus file on disk.
    :rtype: None

    :raises FileNotFoundError: If ``filename`` does not exist.
    :raises ValueError: If array lengths in ``hkl_slice`` (or provided axes) do not match,
        or if the input shapes are incompatible with the expected structure.
    :raises IOError: If the file cannot be opened for read/write.
    :raises RuntimeError: If writing the analysis groups/datasets fails due to
        backend/format issues.

    **Notes**
    - The function assumes the target file already follows the msmapper NeXus layout
      and will append analysis under a dedicated NXgroup (e.g., ``/entry/analysis``).

    **Example**
    ::
        import numpy as np

        H = np.linspace(-1.0, 1.0, 201)
        K = np.zeros_like(H)
        L = np.zeros_like(H)

        # Optional axes (example placeholders)
        ox = H.copy()
        oy = K.copy()
        oz = L.copy()

        update_msmapper_nexus(
            filename="scan_001.nxs",
            hkl_slice=(H, K, L),
            orthogonal_axes=(ox, oy, oz),
            average_axes=None,
            fit_options={"model": "gaussian", "max_iter": 500},
            h_result=h_fit,  # instances of FitResults
            k_result=None,
            l_result=None,
            q_result=q_fit,
            tth_result=None
        )
    """

    h_slice, k_slice, l_slice = hkl_slice
    qx, qy, qz = orthogonal_axes
    wavevector, tth, mag_average = average_axes

    print(f"Updating NeXus file: {filename}")

    with h5py.File(filename, 'a') as hdf:
        entry = nw.add_nxentry(hdf, 'analysis', definition=None, default=True)
        nw.add_nxprocess(entry, 'process', program='mmg_toolbox', version=version_info())

        # hkl
        h = hdf['/processed/reciprocal_space/h-axis'][:]
        k = hdf['/processed/reciprocal_space/k-axis'][:]
        l = hdf['/processed/reciprocal_space/l-axis'][:]
        h_axis = nw.add_nxdata(entry, 'h_axis', ['h'], 'intensity')
        h_axis['h'] = h5py.SoftLink('/processed/reciprocal_space/h-axis')
        nw.add_nxfield(h_axis, 'intensity', h_slice)

        k_axis = nw.add_nxdata(entry, 'k_axis', ['k'], 'intensity')
        k_axis['k'] = h5py.SoftLink('/processed/reciprocal_space/k-axis')
        nw.add_nxfield(k_axis, 'intensity', k_slice)

        l_axis = nw.add_nxdata(entry, 'l_axis', ['l'], 'intensity', default=True)
        l_axis['l'] = h5py.SoftLink('/processed/reciprocal_space/l-axis')
        nw.add_nxfield(l_axis, 'intensity', l_slice)

        # Q
        orthog = nw.add_nxdata(entry, 'orthogonal_axes', ['Qx', 'Qy', 'Qz'], 'volume', 'weight')
        nw.add_nxfield(orthog, 'Qx', qx, units='1/angstrom')
        nw.add_nxfield(orthog, 'Qy', qy, units='1/angstrom')
        nw.add_nxfield(orthog, 'Qz', qz, units='1/angstrom')
        orthog['volume'] = h5py.SoftLink('/processed/reciprocal_space/volume')
        orthog['weight'] = h5py.SoftLink('/processed/reciprocal_space/weight')
        nw.add_attr(orthog, Qx_indices=1, Qy_indices=2, Qz_indices=0, default_slice=len(l) // 2)

        # volume averaged at each magnitude
        qvsi = nw.add_nxdata(entry, 'wavevector', ['q'], 'intensity')
        nw.add_nxfield(qvsi, 'q', wavevector, units='1/angstrom')
        nw.add_nxfield(qvsi, 'intensity', mag_average)

        tthvsi = nw.add_nxdata(entry, 'twotheta', ['tth'], 'intensity')
        nw.add_nxfield(tthvsi, 'tth', tth, units='deg')
        nw.add_nxfield(tthvsi, 'intensity', mag_average)

        if fit_options:
            fit_proc = nw.add_nxprocess(entry, 'peak_fit',
                                        program='mmg_toolbox.utils.fitting.multipeakfit',
                                        **fit_options)
            nw.add_nxnote(fit_proc, 'fit_results_h', 'fit results for h-axis', data=str(h_result))
            nw.add_nxparameters(fit_proc, 'fit_data_h', **h_result.results())
            nw.add_nxnote(fit_proc, 'fit_results_k', 'fit results for k-axis', data=str(k_result))
            nw.add_nxparameters(fit_proc, 'fit_data_k', **k_result.results())
            nw.add_nxnote(fit_proc, 'fit_results_l', 'fit results for l-axis', data=str(l_result))
            nw.add_nxparameters(fit_proc, 'fit_data_l', **l_result.results())
            nw.add_nxnote(fit_proc, 'fit_results_q', 'fit results for wavevector magnitude |Q|', data=str(q_result))
            nw.add_nxparameters(fit_proc, 'fit_data_q', **q_result.results())
            nw.add_nxnote(fit_proc, 'fit_results_tth', 'fit results for remapped two-theta', data=str(tth_result))
            nw.add_nxparameters(fit_proc, 'fit_data_tth', **tth_result.results())

            nw.add_nxfield(h_axis, 'fit', h_result.fit_data(h, ntimes=1)[1], add_to_signal=True)
            nw.add_nxfield(k_axis, 'fit', k_result.fit_data(k, ntimes=1)[1], add_to_signal=True)
            nw.add_nxfield(l_axis, 'fit', l_result.fit_data(l, ntimes=1)[1], add_to_signal=True)
            nw.add_nxfield(qvsi, 'fit', q_result.fit_data(wavevector, ntimes=1)[1], add_to_signal=True)
            nw.add_nxfield(tthvsi, 'fit', tth_result.fit_data(tth, ntimes=1)[1], add_to_signal=True)




"""
SpectraContainer object

=== DATA MODEL ===
spectra = Spectra(energy, signal, mode, process)
metadata = XasMetadata(scan_no=1234, default_mode='tey', sample_name='Fe')
scan = SpectraContainer('name', {'mode': spectra}, metadata=metadata)
scan2 = scan + 2  # add 2 to signal of each contained mode
scan.remove_background()  # apply operation to each contained mode, store previous version in scan.parents
"""
from __future__ import annotations

import inspect
from functools import wraps
import datetime
import numpy as np
import h5py
import matplotlib.pyplot as plt

from mmg_toolbox.nexus import nexus_writer as nw
from . import spectra_analysis as spa
from mmg_toolbox.utils.polarisation import pol_subtraction_label
from .spectra import Spectra, SpectraSubtraction


class Metadata:
    filename: str = ''
    beamline: str = ''
    scan_no: int = 0
    start_date_iso: str = ''
    end_date_iso: str = ''
    cmd: str = ''
    pol: str = 'pc'
    sample_name: str = ''
    temp: float = 300
    mag_field: float = 0
    pitch: float = 0  # 0 == sample surface normal to beam

    def __init__(self, **kwargs):
        for name, value in kwargs.items():
            if hasattr(self, name):
                setattr(self, name, value)

    def __str__(self):
        return str(self.__dict__)


class XasMetadata(Metadata):
    default_mode: str = 'tey'
    element: str = ''
    edge: str = ''
    energy: np.ndarray = np.arange(10)
    monitor: np.ndarray = np.ones(10)
    raw_signals: dict[str, np.ndarray] = {'tey': np.zeros(10)}


def spectra_method_decorator(target_cls):
    """Add methods from Spectra to SpectraContainer"""
    for name, method in inspect.getmembers(Spectra, predicate=inspect.isfunction):
        if name in ['divide_by_signal_at_energy', 'divide_by_preedge', 'divide_by_postedge', 'norm_to_peak',
                    'norm_to_jump', 'remove_background', 'auto_edge_background']:
            @wraps(method)
            def fn(self, *args, _method=method, **kwargs):
                self.parents = (self.copy(), *self.parents)
                self.spectra = {n: _method(s, *args, **kwargs) for n, s in self.spectra.items()}
                self.process_label = next(iter(self.spectra.values())).process_label
            setattr(target_cls, name, fn)

        elif name in ['create_nxnote', 'create_nxdata', 'plot', 'plot_bkg', 'plot_parents']:
            @wraps(method)
            def fn(self, *args, _method=method, **kwargs):
                return [_method(s, *args, **kwargs) for s in self.spectra.values()]
            setattr(target_cls, name, fn)

    return target_cls


@spectra_method_decorator
class SpectraContainer:
    """
    Container for Spectra and metadata
    """

    def __init__(self, name: str, spectra: dict[str, Spectra | SpectraSubtraction],
                 *parents: SpectraContainer, metadata: XasMetadata = XasMetadata()):
        self.name = name
        self.process_label = next(iter(spectra.values())).process_label
        self.parents = parents
        self.spectra = spectra
        self.metadata = metadata

    def __repr__(self):
        return f"SpectraContainer('{self.name}', '{self.process_label}', {list(self.spectra)})"

    def __str__(self):
        meta_str = (
                f"{self.metadata.filename}\n" +
                f"{self.metadata.start_date_iso}\n" +
                f"{self.metadata.cmd}\n" +
                f"mode: '{self.metadata.default_mode}', signals: {list(self.spectra)}\n" +
                f"E = {np.mean(self.metadata.energy):.2f} eV -> {self.metadata.element} {self.metadata.edge}\n" +
                f"   Sample: '{self.metadata.sample_name}'\n" +
                f"T = {self.metadata.temp:.2f} K\n" +
                f"B = {self.metadata.mag_field:.2f} T\n" +
                f"Pol = '{self.metadata.pol}'"
        )
        return meta_str

    def __iter__(self):
        return self.spectra.__iter__()

    def __getitem__(self, item):
        return self.spectra[item]

    def __add__(self, other):
        if issubclass(type(other), SpectraContainer):
            # average Spectra
            spectra = {n: s + other.spectra[n] for n, s in self.spectra.items() if n in other.spectra}
        else:
            # add float or array to Spectra
            spectra = {n: s + other for n, s in self.spectra.items()}
        return SpectraContainer(self.name, spectra, self, *self.parents, metadata=self.metadata)

    def __sub__(self, other):
        if issubclass(type(other), SpectraContainer):
            # Subtract Spectra
            return SpectraContainerSubtraction(self, other)
        else:
            # Subtract float or array
            spectra = {n: s - other for n, s in self.spectra.items()}
            return SpectraContainer(self.name, spectra, self, *self.parents, metadata=self.metadata)

    def __mul__(self, other):
        if issubclass(type(other), SpectraContainer):
            raise Exception('Cannot multiply SpectraContainer')
        else:
            # multiply Spectra by float or array
            spectra = {n: s * other for n, s in self.spectra.items()}
            return SpectraContainer(self.name, spectra, self, *self.parents, metadata=self.metadata)

    def copy(self, name=None):
        """Create copy of spectra container using new name"""
        name = name or self.name
        return SpectraContainer(name, self.spectra, *self.parents, metadata=self.metadata)

    def label(self):
        # return f"{self.name} {self.process_label}"
        return self.process_label.replace('/', '').replace(' ', '')

    def analysis_steps(self):
        return {sc.label(): sc.spectra for sc in list(reversed(self.parents)) + [self]}

    def nx_entry(self, nexus: h5py.File, name='entry', default=True) -> h5py.Group:
        entry = nw.add_nxentry(nexus, name, definition='NXxas')
        nw.add_nxfield(entry, 'entry_identifier', self.metadata.scan_no)
        nw.add_nxfield(entry, 'start_time', self.metadata.start_date_iso)
        nw.add_nxfield(entry, 'end_time', self.metadata.end_date_iso)
        nw.add_nxfield(entry, 'scan_command', self.metadata.cmd)
        nw.add_nxfield(entry, 'mode', self.metadata.default_mode)
        nw.add_nxfield(entry, 'element', self.metadata.element)
        nw.add_nxfield(entry, 'edge', self.metadata.edge)
        nw.add_nxfield(entry, 'polarization_label', self.metadata.pol)
        if default:
            nexus.attrs['default'] = name
        return entry

    def nx_instrument(self, entry: h5py.Group) -> h5py.Group:
        energy = self.metadata.energy
        monitor = self.metadata.monitor
        raw_signals = self.metadata.raw_signals
        mode = self.metadata.default_mode

        instrument = nw.add_nxinstrument(root=entry, name='instrument', instrument_name=self.metadata.beamline)
        nw.add_nxsource(instrument, 'source')
        nw.add_nxmono(instrument, 'mono', energy_ev=energy)
        nw.add_nxdetector(instrument, 'incoming_beam', data=monitor)
        nw.add_nxdetector(instrument, 'absorbed_beam', data=raw_signals[mode])
        for name, signal in raw_signals.items():
            nw.add_nxdetector(instrument, name, data=signal)
        return instrument

    def nx_sample(self, entry: h5py.Group) -> h5py.Group:
        sample = nw.add_nxsample(
            root=entry,
            name='sample',
            sample_name=self.metadata.sample_name,
            chemical_formula='',
            temperature_k=self.metadata.temp,
            magnetic_field_t=self.metadata.mag_field,
            electric_field_v=0,
            mag_field_dir='z',
            electric_field_dir='z',
            sample_type='sample',
            description=''
        )
        energy = self.metadata.energy
        nw.add_nxbeam(
            root=sample,
            name='beam',
            incident_energy_ev=float(np.mean(energy)),
            polarisation_label=self.metadata.pol,
            beam_size_um=None
        )
        return sample

    def nx_process(self, entry: h5py.Group) -> h5py.Group:
        from mmg_toolbox import __version__

        # NXprocess - read dat
        input_filename = self.metadata.filename
        if input_filename.endswith('.dat'):
            read_dat = nw.add_nxprocess(
                root=entry,
                name='read_dat',
                program='xmcd_analysis_functions',
                version=__version__,
                date=str(datetime.datetime.now()),
                sequence_index=1,
            )
            nw.add_nxnote(
                root=read_dat,
                name='dat_file',
                data=open(input_filename, 'r').read(),
                filename=input_filename,
                description='DLS SRS format',
                sequence_index=1
            )

        # NXProcess
        process = nw.add_nxprocess(
            root=entry,
            name='process',
            program='mmg_toolbox',
            version=__version__,
            date=str(datetime.datetime.now()),
            sequence_index=2 if self.metadata.filename.endswith('.dat') else 1,
        )
        return process

    def nx_analysis_steps(self, entry: h5py.Group, process: h5py.Group):
        analysis_steps = self.analysis_steps()
        for n, (name, spectra) in enumerate(analysis_steps.items()):
            spectra[self.metadata.default_mode].create_nxnote(process, name, n + 1)

        # NXdata groups
        for name, spectra in analysis_steps.items():
            mode_spectra = spectra[self.metadata.default_mode]
            data = mode_spectra.create_nxdata(entry, name, default=True)
            aux_signals = []
            for signal, spec in spectra.items():
                nw.add_nxfield(data, signal, spec.signal, units='')
                aux_signals.append(signal)
                if spec.background is not None:
                    name = f"{signal}_background"
                    nw.add_nxfield(data, name, spec.background, units='')
                    aux_signals.append(name)
            data.attrs['auxiliary_signals'] = aux_signals

    def nx_main_entry(self, nexus: h5py.File, name='entry', default=True):
        entry = self.nx_entry(nexus, name=name, default=default)
        self.nx_instrument(entry)
        self.nx_sample(entry)
        process = self.nx_process(entry)
        self.nx_analysis_steps(entry, process)

    def _nx_add_items(self, nexus: h5py.File):
        nw.add_entry_links(nexus, self.metadata.filename)
        self.nx_main_entry(nexus)

    def write_nexus(self, nexus_filename: str):
        with h5py.File(nexus_filename, 'w') as nxs:
            self._nx_add_items(nxs)
        print(f'Created {nexus_filename}')

    def create_figure(self):
        fig, axs = plt.subplots(1, len(self.spectra))

        for ax, s in zip(axs, self.spectra.values()):
            s.plot(ax)
            ax.set_xlabel('E [eV]')
            ax.set_ylabel('signal')
            ax.legend()
        return fig


class SpectraContainerSubtraction(SpectraContainer):
    """Special subclass for subtraction of SpectraContainers - XMCD and XMLD"""
    def __init__(self, spectra_container1: SpectraContainer, spectra_container2: SpectraContainer):
        # subtract each spectra in container
        spectra = {
            name: spectra - spectra_container2.spectra[name]
            for name, spectra in spectra_container1.spectra.items()
            if name in spectra_container2.spectra
        }
        # subtraction name
        if spectra_container1.metadata.pol != spectra_container2.metadata.pol:
            name = pol_subtraction_label(spectra_container1.metadata.pol)
            # rename parents (for display)
            spectra_container1 = spectra_container1.copy(spectra_container1.metadata.pol)
            spectra_container2 = spectra_container2.copy(spectra_container2.metadata.pol)
        else:
            name = 'subtraction'
        # subtraction metadata (merge these?)
        metadata = XasMetadata(**spectra_container1.metadata.__dict__)
        metadata.filename = ''
        super().__init__(name, spectra, spectra_container1, spectra_container2, metadata=metadata)

    def nx_sum_rules_process(self, entry: h5py.Group):
        from mmg_toolbox import __version__
        process = nw.add_nxprocess(
            root=entry,
            name='sum_rules',
            program='mmg_toolbox',
            version=__version__,
            date=str(datetime.datetime.now()),
            sequence_index=2,
        )
        try:
            n_holes = spa.default_n_holes(self.metadata.element)
        except KeyError as ke:
            print(f"Warning: {ke}")
            n_holes = 1
        for n, (name, spectra) in enumerate(self.spectra.items()):
            spectra.create_sum_rules_nxnote(n_holes, process, name, n + 1, element=self.metadata.element)

    def _nx_add_items(self, nexus: h5py.File):
        for parent in self.parents:
            parent.nx_main_entry(nexus, name=parent.name, default=False)
        entry = self.nx_entry(nexus, name=self.name, default=True)
        self.nx_sample(entry)
        process = self.nx_process(entry)
        self.nx_sum_rules_process(entry)
        self.nx_analysis_steps(entry, process)



def average_polarised_scans(*scans: SpectraContainer) -> list[SpectraContainer]:
    """Find unique polarisations and average each scan at that polarisation"""
    pol_scans = {
        pol: [scan for scan in scans if scan.metadata.pol == pol]
        for pol in {scan.metadata.pol for scan in scans}
    }
    average_scans = {
        pol: sum(scan_list[1:], scan_list[0]) if len(scan_list) > 1 else scan_list[0]
        for pol, scan_list in pol_scans.items()
    }
    # rename containers
    for pol, scan in average_scans.items():
        scan.name = pol
        for spectra in scan.spectra.values():
            spectra.process_label += f"_{pol}"
    return list(average_scans.values())




"""
tk widget for editing the Config file
"""
import os

from ..misc.styles import tk, ttk, create_root
from ..misc.logging import create_logger
from ..misc.functions import topmenu
from ..misc.config import get_config, save_config, default_config, C
from ..misc.matplotlib import COLORMAPS
from ..widgets.roi_editor import RoiEditor
from .edit_text import EditText

logger = create_logger(__file__)


class ConfigEditor:
    """
    Edit the Configuration File in an inset window
    """

    def __init__(self, parent: tk.Misc, config: dict | None = None):
        self.root = create_root('Config. Editor', parent)
        # self.root.wm_overrideredirect(True)
        self.config = config or get_config()
        self.config_setters = {}
        self.config_getters = {}

        menu = {
            'Config': {
                'Reset': self.reset_config,
                'Delete config file': self.delete_config,
                'View config file': self.view_config,
            }
        }

        topmenu(self.root, menu)

        self.window = ttk.Frame(self.root, borderwidth=20, relief=tk.RAISED)
        self.window.pack(side=tk.TOP, fill=tk.BOTH)

        frm = ttk.Frame(self.window)
        frm.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)
        var = ttk.Label(frm, text='Edit Config. Parameters', style="title.TLabel")
        var.pack(expand=tk.YES, fill=tk.X, padx=10, pady=10)

        # parameter entry boxes
        self.create_param(C.conf_file, 'Config File:')
        self.create_param(C.beamline, 'Beamline:')
        self.create_param(C.normalise_factor, 'Normalise:')
        self.create_tuple_param(C.text_size, 'Text Size:', 'chars x rows')
        self.create_tuple_param(C.plot_size, 'Plot Size:', 'w x h inches')
        self.create_tuple_param(C.image_size, 'Image Size:', 'w x h inches')
        self.create_tuple_param(C.plot_max_percent, 'Max Plot Size:', 'w x h % of screen')
        self.create_param(C.plot_dpi, 'Figure DPI:')
        self.create_list_param(C.default_colormap, 'Default colormap:', *COLORMAPS)
        self.create_param(C.metadata_label, 'Metadata label', button=self.metadata_list_window)
        self.create_text_param(C.metadata_string, 'Metadata expression')

        # Buttons at bottom
        frm = ttk.Frame(self.window)
        frm.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)
        var = ttk.Button(frm, text='ROIs', command=self.roi_window)
        var.pack(side=tk.LEFT, expand=tk.YES)
        var = ttk.Button(frm, text='Save', command=self.save_config)
        var.pack(side=tk.LEFT, expand=tk.YES)
        var = ttk.Button(frm, text='Update', command=self.update_config)
        var.pack(side=tk.LEFT, fill=tk.X, expand=tk.YES)

    def create_param(self, config_name: str, label: str, button=None):
        variable = tk.StringVar(self.root, self.config.get(config_name, ''))
        get_type = type(self.config.get(config_name, ''))
        self.config_setters[config_name] = lambda name=config_name: str(self.config.get(name, ''))
        self.config_getters[config_name] = lambda: get_type(variable.get())

        frm = ttk.Frame(self.window)
        frm.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH, padx=10, pady=5)
        ttk.Label(frm, text=label, width=20).pack(side=tk.LEFT, padx=2)
        ttk.Entry(frm, textvariable=variable).pack(side=tk.LEFT, fill=tk.X, expand=tk.YES)
        if button is not None:
            ttk.Button(frm, text='...', command=button, width=3).pack(side=tk.LEFT)

    def create_tuple_param(self, config_name: str, label: str, units: str = ''):
        values = self.config.get(config_name, ())
        types = [type(val) for val in values]
        tuple_vars = tuple(
            tk.StringVar(self.root, str(val))
            for val in values
        )

        def setter(name=config_name, _vars=tuple_vars):
            _values = self.config.get(name, ())
            for _var, _val in zip(_vars, _values):
                _var.set(str(_val))

        def getter():
            return tuple([
                var_type(_var.get())
                for _var, var_type in zip(tuple_vars, types)
            ])

        frm = ttk.Frame(self.window)
        frm.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH, padx=10, pady=5)
        ttk.Label(frm, text=label, width=20).pack(side=tk.LEFT, padx=2)
        for var in tuple_vars:
            ttk.Entry(frm, textvariable=var, width=4).pack(side=tk.LEFT)
        ttk.Label(frm, text=units).pack(side=tk.LEFT, fill=tk.X)

        self.config_setters[config_name] = setter
        self.config_getters[config_name] = getter
        setter()

    def create_list_param(self, config_name: str, label: str, *list_names: str):
        frm = ttk.Frame(self.window)
        frm.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH, padx=10, pady=5)
        ttk.Label(frm, text=label, width=20).pack(side=tk.LEFT, padx=2)
        default = self.config.get(config_name, list_names[0])
        var = tk.StringVar(self.root, default)
        self.config_getters[config_name] = var.get
        self.config_setters[config_name] = lambda name=config_name: var.set(self.config.get(name, ''))
        var2 = tk.StringVar(self.root, default)
        cbox = ttk.Combobox(frm, textvariable=var2, values=list_names)
        cbox.pack(side=tk.LEFT)
        cbox.bind('<<ComboboxSelected>>', lambda e: var.set(var2.get()))

    def create_text_param(self, config_name: str, label: str):

        def button(name=config_name):
            text = self.config.get(name, '')
            new_text = EditText(
                expression=text,
                parent=self.root,
                textwidth=self.config.get(C.text_size, (50, 20))[0]
            ).show()
            if new_text:
                self.config[name] = new_text

        frm = ttk.Frame(self.window)
        frm.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH, padx=10, pady=5)
        ttk.Label(frm, text=label, width=20).pack(side=tk.LEFT, padx=2)
        ttk.Button(frm, text='Edit', command=button, width=3).pack(side=tk.LEFT)

    def _update_config(self):
        updated_config = {
            name: var() for name, var in self.config_getters.items()
        }
        self.config.update(updated_config)

    def update_config(self):
        self._update_config()
        self.root.destroy()

    def set_from_config(self, config: dict):
        self.config.update(config)
        for name, setter in self.config_setters.items():
            print(name, self.config[name])
            setter()

    def reset_config(self):
        beamline = self.config_getters[C.beamline]()
        default = default_config(beamline)
        self.set_from_config(default)

    def delete_config(self):
        config_file = self.config.get(C.conf_file)
        if config_file and os.path.isfile(config_file):
            from tkinter import messagebox
            answer = messagebox.askyesno(
                title='Config',
                message='Delete config file?\n {}'.format(config_file),
                parent=self.root
            )
            if answer:
                os.remove(config_file)
                print('Removed config file')

    def save_config(self):
        self._update_config()
        save_config(self.config)
        self.root.destroy()

    def view_config(self):
        config_file = self.config.get(C.conf_file)
        if config_file and os.path.isfile(config_file):
            from .edit_text import EditText
            import json

            dump = json.dumps(json.load(open(config_file, 'r')), sort_keys=False, indent=4)
            EditText(dump, self.root).show()

    def roi_window(self):
        window = create_root('Regions of Interest (ROIs)', self.root)
        RoiEditor(window, self.config)

    def metadata_list_window(self):
        MetadataListEditor(self.root, self.config)


class MetadataListEditor:
    """
    Edit the metadata list
    """

    def __init__(self, parent: tk.Misc, config: dict | None = None):
        self.root = create_root('Metadata List', parent)
        # self.root.wm_overrideredirect(True)
        self.param_list = []

        if config is None:
            self.config = get_config()
        else:
            self.config = config

        metadata_list = self.config.get(C.metadata_list, {})
        self.window = ttk.Frame(self.root)
        self.window.pack(fill=tk.BOTH, expand=tk.YES)
        for name, expression in metadata_list.items():
            self.create_entry(name, expression)
        self.create_entry('', '')

        frm = ttk.Frame(self.root)
        frm.pack(side=tk.TOP, anchor=tk.W, padx=5)
        ttk.Button(frm, text='+', command=self.create_entry, width=4).pack(side=tk.LEFT)

        frm = ttk.Frame(self.root)
        frm.pack(side=tk.TOP, fill=tk.BOTH, expand=tk.YES)
        ttk.Button(frm, text='Update', command=self.update).pack(fill=tk.X, expand=tk.YES)

    def create_entry(self, name: str = '', expression: str = ''):
        name_var = tk.StringVar(self.root, name)
        expression_var = tk.StringVar(self.root, expression)
        self.param_list.append((name_var, expression_var))

        frm = ttk.Frame(self.window)
        frm.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH, padx=10, pady=5)
        ttk.Entry(frm, textvariable=name_var, width=10).pack(side=tk.LEFT, fill=tk.X, expand=tk.YES)
        ttk.Entry(frm, textvariable=expression_var, width=40).pack(side=tk.LEFT, fill=tk.X, expand=tk.YES)

    def update(self):
        metadata_list = {
            name.get(): expression.get() for name, expression in self.param_list
            if name.get() and expression.get()
        }
        self.config[C.metadata_list] = metadata_list
        self.root.destroy()




"""
Script & Notebook templates
"""

import os
import re
import datetime

re_replacement = re.compile(r"{{(.+?)}}")


class R:
    """Names used in replacements"""
    beamline = 'beamline'
    description = 'description'
    filepaths = 'filepaths'
    exp = 'experiment_dir'
    proc = 'processing_dir'
    scannos = 'scan_numbers'
    title = 'title'
    xaxis = 'x-axis'
    yaxis = 'y-axis'
    value = 'value'

SCRIPTS = {
    # name: (filename, description)
    'example': ('example_script.py', 'a simple example'),
    'plot multi-line': ('experiment_multiline.py', 'create a multi-line plot'),
    'peak fitting': ('experiment_fitting.py', 'fit peaks and plot the results'),
    'spectra': ('spectra_script.py', 'normalise spectra and subtract polarisations')
}

NOTEBOOKS = {
    # name: (filename, description)
    'example': ('example_notebook.ipynb', 'a basic example'),
}

TEMPLATE = {
    # {{template}}: replacement
    R.description: 'a short description',
    R.filepaths: "'file1.nxs', 'file2.nxs', 'file3.nxs'",
    R.exp: 'path/to/dir',
    R.scannos: 'range(-10, 0)',
    R.title: 'a nice plot',
    R.xaxis: 'axes',
    R.yaxis: 'signal',
    R.value: 'Ta'
}


def find_replacements(string: str) -> list[str]:
    """find all replacements in string"""
    return re_replacement.findall(string)


def generate_script(template_name: str, **replacements) -> str:
    """generate script str from template"""
    template_file, description = SCRIPTS[template_name]
    template_file = os.path.join(os.path.dirname(__file__), template_file)
    template_changes = TEMPLATE.copy()
    template_changes.update(replacements)
    template_changes['date'] = str(datetime.date.today())

    template_string = open(template_file, 'r').read()
    for name, value in template_changes.items():
        param = "{{" + name + "}}"
        print(f"Replacing {template_string.count(param)} instances of {param}")
        template_string = template_string.replace(param, value)
    return template_string


def create_script(new_script_path: str, template_name: str, **replacements):
    """create script from template"""
    script = generate_script(template_name, **replacements)

    with open(new_script_path, 'w') as new:
        new.write(script)
    print(f"Created {new_script_path}")


def create_notebook(new_notebook_path: str, template_name: str, **replacements):
    """create script from template"""
    template_file, description = NOTEBOOKS[template_name]
    template_file = os.path.join(os.path.dirname(__file__), template_file)
    template_changes = TEMPLATE.copy()
    template_changes.update(replacements)

    template_string = open(template_file, 'r').read()
    for name, value in template_changes.items():
        param = "{{" + name + "}}"
        print(f"Replacing {template_string.count(param)} instances of {param}")
        template_string = template_string.replace(param, value)

    with open(new_notebook_path, 'w') as new:
        new.write(template_string)
    print(f"Created {new_notebook_path}")




"""
Matplotlib plotting functions

e.g.
>> set_plot_defaults()
>> axs = create_multiplot(2, 2, title='New figure')
>> plot_line(axs[0], x, y)
>> plot_line(axs[1], x2, y2, label='data 2')
>> plot_lines(ax[2], x, [y, y2])
>> plot_detector_image(ax[3], image)
>> labels('title', 'x', 'y', legend=True, axes=axs[0])
"""

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.pyplot import Axes, Line2D
from matplotlib.collections import QuadMesh
from matplotlib.patches import Polygon, Rectangle
from mpl_toolkits.mplot3d.axes3d import Axes3D
from mpl_toolkits.mplot3d.art3d import Poly3DCollection

DEFAULT_FONTSIZE = 14
FIG_SIZE = [12, 8]
FIG_DPI = 80
DEFAULT_CMAP = 'viridis'


'----------------------------Plot manipulation--------------------------'


def set_plot_defaults(rcdefaults=False):
    """
    Set custom matplotlib rcparams, or revert to matplotlib defaults
    These handle the default look of matplotlib plots
    See: https://matplotlib.org/stable/tutorials/introductory/customizing.html#the-default-matplotlibrc-file
    :param rcdefaults: False*/ True, if True, revert to matplotlib defaults
    :return: None
    """
    if rcdefaults:
        print('Return matplotlib rcparams to default settings.')
        plt.rcdefaults()
        return

    plt.rc('figure', figsize=FIG_SIZE, dpi=FIG_DPI, autolayout=False)
    plt.rc('lines', marker='o', color='r', linewidth=2, markersize=6)
    plt.rc('errorbar', capsize=2)
    plt.rc('legend', loc='best', frameon=False, fontsize=DEFAULT_FONTSIZE)
    plt.rc('axes', linewidth=2, titleweight='bold', labelsize='large')
    plt.rc('xtick', labelsize='large')
    plt.rc('ytick', labelsize='large')
    plt.rc('axes.formatter', limits=(-3, 3), offset_threshold=6)
    plt.rc('image', cmap=DEFAULT_CMAP)  # default colourmap, see https://matplotlib.org/stable/gallery/color/colormap_reference.html
    # Note font values appear to only be set when plt.show is called
    plt.rc(
        'font',
        family='serif',
        style='normal',
        weight='bold',
        size=DEFAULT_FONTSIZE,
        serif=['Times New Roman', 'Times', 'DejaVu Serif']
    )
    # plt.rcParams["savefig.directory"] = os.path.dirname(__file__) # Default save directory for figures


def generate_subplots(n_plots: int, subplots: tuple[int, int] = (4, 4), ax_size: tuple[float, float] = (5, 3),
                      suptitle: str | None = None) -> list[tuple[plt.Figure, Axes]]:
    """
    Generate a grid of n_plots.
    If n_plots is larger than the number of subplots on a figure, multiple figures will be generated.

        fig_ax = generate_subplots(25, subplots=(4, 4), suptitle='plots')
        for fig, ax in fig_ax:
            ax.plot(x, y)

    :param n_plots: number of subplots to generate
    :param subplots: [nrows, ncols] number of subplots to generate per figure [vertical, horizontal]
    :param ax_size: [horiz, vert] size of each axis in inches (scaled by dpi)
    :param suptitle: title of each figure (the figure number will be appended)
    :return: list of (fig, axes), length n_plots
    """
    nrows, ncols = subplots
    n_figures = int(np.ceil(n_plots / float(nrows * ncols)))
    hsize, vsize = ax_size

    fig_ax = []
    for n in range(n_figures):
        fig, axs = plt.subplots(nrows, ncols, figsize=[hsize * ncols, vsize * nrows], dpi=FIG_DPI)
        fig.subplots_adjust(hspace=0.35, wspace=0.32, left=0.07, right=0.97)
        if suptitle is not None:
            fig.suptitle(suptitle + f" {n+1}/{n_figures}" if n_figures > 1 else "")
        ax_list = [(fig, ax) for ax in axs.flatten()]
        for fig, ax in ax_list[n_plots:]:
            ax.set_axis_off()
        fig_ax.extend(ax_list[:n_plots])
        n_plots -= len(ax_list)
    return fig_ax


def plot_line(axes: Axes, xdata: np.ndarray, ydata: np.ndarray, yerrors: np.ndarray | None = None,
              line_spec: str = '-o', *args, **kwargs) -> list[Line2D]:
    """
    Plot line on given matplotlib axes subplot
    Uses matplotlib.plot or matplotlib.errorbar if yerrors is not None
    :param axes: matplotlib figure or subplot axes, None uses current axes
    :param xdata: array data on x axis
    :param ydata: array data on y axis
    :param yerrors: array errors on y axis (or None)
    :param line_spec: str matplotlib.plot line_spec
    :param args: additional arguments
    :param kwargs: additional arguments
    :return: output of plt.plot [line], or plt.errorbar [line, xerrors, yerrors]
    """
    if yerrors is None:
        lines = axes.plot(xdata, ydata, line_spec, *args, **kwargs)
    else:
        lines = axes.errorbar(xdata, ydata, yerrors, *args, fmt=line_spec, **kwargs)
    return lines


PlotData = tuple[float | None, np.ndarray, np.ndarray, np.ndarray] | tuple[float | None, np.ndarray, np.ndarray]

def plot_lines(axes: Axes, *plot_data: PlotData,
               cmap: str = DEFAULT_CMAP, line_spec: str = '-o', **kwargs) -> tuple[list[Line2D], plt.cm.ScalarMappable]:
    """
    Plot lines on given matplotlib axes subplot
    Uses matplotlib.plot or matplotlib.errorbar if yerrors is not None
    :param axes: matplotlib figure or subplot axes, None uses current axes
    :param plot_data: [value, xdata, ydata, yerrors] or [value, xdata, ydata]
    :param cmap: name of colormap to generate colour variation in lines
    :param line_spec: str or list[m] of str matplotlib.plot line_spec
    :param kwargs: additional arguments
    :return: output of plt.plot [line], or plt.errorbar [line, xerrors, yerrors]
    :return: ScalarMappable for use in colorbar
    """

    cdata = np.array([data[0] for data in plot_data])
    if None in cdata:
        cdata = np.arange(len(plot_data))
    norm = plt.Normalize()
    cnorm = norm(cdata)
    # cnorm = cdata - (cdata.min() - max(0.05 * ))
    # cnorm = cnorm / cnorm.max()
    cols = plt.get_cmap(cmap)(cnorm)

    # Create ScalarMappable for colorbar
    sm = plt.cm.ScalarMappable(cmap=cmap)
    sm.set_array(cdata)

    lines = []
    for data, col in zip(plot_data, cols):
        value, x, y, err = (data + (None, ) * 3)[:4]
        if y is None:
            y = x
            x = np.arange(y)
        lines.extend(plot_line(axes, x, y, err, line_spec, c=col, **kwargs))
    return lines, sm


def plot_image(axes: Axes, image: np.ndarray, clim: tuple[float, float] = None,
               *args, **kwargs) -> Axes:
    """
    Plot 2D image
    :param axes: matplotlib figure or subplot axes, None uses current axe
    :param image: 2d array image data
    :param clim: None or [min, max] values for color cutoff
    :param args: additional arguments for plt.pcolormesh
    :param kwargs: additional arguments for plt.pcolormesh
    :return: axes object
    """

    if 'shading' not in kwargs.keys():
        kwargs['shading'] = 'gouraud'
    if clim:
        kwargs['vmin'] = clim[0]
        kwargs['vmax'] = clim[1]

    axes.pcolormesh(image, *args, **kwargs)
    axes.invert_yaxis()
    axes.axis('image')
    return axes


def plot_2d_surface(axes: Axes, image: np.ndarray,
                    xdata: np.ndarray | None = None, ydata: np.ndarray = None,
                    clim: tuple[float, float] = None, axlim: str = 'image', **kwargs) -> QuadMesh:
    """
    Plot 2D data as colourmap surface
    :param axes: matplotlib figure or subplot axes, None uses current axe
    :param image: 2d array image data
    :param xdata: array data, 2d or 1d
    :param ydata: array data 2d or 1d
    :param clim: None or [min, max] values for color cutoff from plt.clim
    :param axlim: axis limits from plt.axis
    :param kwargs: additional arguments for plt.pcolormesh
    :return: output of plt.pcolormesh
    """

    if 'shading' not in kwargs:
        kwargs['shading'] = 'gouraud'
    if clim:
        kwargs['vmin'] = clim[0]
        kwargs['vmax'] = clim[1]

    if xdata is None or ydata is None:
        surface = axes.pcolormesh(image, **kwargs)
    else:
        if np.ndim(xdata) == 1 and np.ndim(ydata) == 1:
            ydata, xdata = np.meshgrid(ydata, xdata)
        surface = axes.pcolormesh(xdata, ydata, image, **kwargs)
    axes.axis(axlim)
    return surface


def plot_3d_lines(axes: Axes3D, zdata: list[np.ndarray],
                  xdata: list[np.ndarray] | None = None, ydata: list[np.ndarray] | None = None,
                  labels: list[str] | None = None, **kwargs) -> list[Line2D]:
    """
    Plot 2D image data as 3d surface
    :param axes: matplotlib figure or subplot axes, None uses current axe
    :param zdata: 2d array [n, m]
    :param xdata: array data, 2d or 1d
    :param ydata: array data 2d or 1d
    :param labels: list of labels for each line, or None
    :param kwargs: additional arguments for plt.plot
    :return: output of plt.plot
    """
    lines = []
    for n, z in enumerate(zdata):
        x = np.arange(z.shape[0]) if xdata is None else xdata[n]
        y = n * np.ones(z.shape[0]) if ydata is None else ydata[n]
        if y.size == 1:
            y = y * np.ones(z.shape[0])
        if labels is not None:
            kwargs['label'] = labels[n]
        lines.extend(axes.plot(x, y, z, **kwargs))
    return lines


def plot_3d_surface(axes: Axes3D, image: np.ndarray,
                    xdata: np.ndarray | None = None, ydata: np.ndarray | None = None,
                    samples: int | None = None, clim: tuple[int, int] | None = None,
                    axlim: str = 'auto', **kwargs) -> Poly3DCollection:
    """
    Plot 2D image data as 3d surface
    :param axes: matplotlib figure or subplot axes, None uses current axe
    :param image: 2d array image data
    :param xdata: array data, 2d or 1d
    :param ydata: array data 2d or 1d
    :param samples: max number of points to take in each direction, by default does not downsample
    :param clim: None or [min, max] values for color cutoff from plt.clim
    :param axlim: axis limits from plt.axis
    :param kwargs: additional arguments for plt.plot_surface
    :return: output of plt.plot_surface
    """
    if samples:
        kwargs['rcount'] = samples
        kwargs['ccount'] = samples
    else:
        # default in plot_surface is 50
        kwargs['rcount'],  kwargs['ccount'] = np.shape(image)
    if clim:
        kwargs['vmin'] = clim[0]
        kwargs['vmax'] = clim[1]

    if xdata is None or ydata is None:
        surface = axes.plot_surface(image, **kwargs)
    else:
        if np.ndim(xdata) == 1 and np.ndim(ydata) == 1:
            ydata, xdata = np.meshgrid(ydata, xdata)
        surface = axes.plot_surface(xdata, ydata, image, **kwargs)
    axes.axis(axlim)
    return surface



def set_span_bounds(span: Rectangle | Polygon, xmin: float, xmax: float, ymin: float, ymax: float):
    """Set bounds for span=ax.axvspan, working for old matplotlib versions"""
    if hasattr(span, "set_bounds"):
        # Rectangle patch
        width = xmax - xmin
        height = ymax - ymin
        span.set_bounds(xmin, ymin, width, height)
    else:
        # Polygon patch: update vertices
        new_verts = [
            (xmin, ymin),
            (xmin, ymax),
            (xmax, ymax),
            (xmax, ymin),
            (xmin, ymin),
        ]
        span.set_xy(new_verts)



"""
test user interface
"""

import numpy as np
from mmg_toolbox.tkguis.misc.styles import create_root
# from mmg_toolbox.tkguis.widgets.python_editor import PythonEditor
# # from mmg_toolbox.tkguis.widgets.folder_treeview import NexusFolderTreeViewFrame
# from mmg_toolbox.tkguis.widgets.nexus_treeview import HDFViewer
# from mmg_toolbox.tkguis.widgets.simple_plot import NexusDefaultPlot
from mmg_toolbox.tkguis import create_nexus_file_browser, create_data_viewer, create_title_window
from mmg_toolbox.tkguis.widgets.scan_selector import FolderScanSelector
# from mmg_toolbox.tkguis.widgets.nexus_details import NexusDetails
from mmg_toolbox.tkguis.widgets.nexus_image import NexusDetectorImage
from mmg_toolbox.tkguis.misc.logging import set_all_logging_level
from mmg_toolbox.tkguis.misc.matplotlib import ini_image

if __name__ == '__main__':
    pass
    # set_all_logging_level('debug')
    # window = create_file_browser()

    # f = "/scratch/grp66007/data/i16/das_example_data/1041304.nxs"
    # f = r"C:\Users\grp66007\OneDrive - Diamond Light Source Ltd\I16\Nexus_Format\example_nexus\1041304.nxs"
    # root = create_root('details')
    # NexusDetails(root, hdf_filename=f)
    # root.mainloop()

    # # HDFViewer(f)
    # obj = NexusDefaultPlot(f)
    # obj.root.mainloop()

    # root = create_root('folder')
    # root.geometry('200x300')
    # FolderScanSelector(root)
    # root.mainloop()

    # create_data_viewer(r"C:\Users\grp66007\OneDrive - Diamond Light Source Ltd\I16\Nexus_Format\example_nexus")
    # create_title_window()

    # root = create_root('image')
    # NexusDetectorImage(root, hdf_filename=r"C:\Users\grp66007\OneDrive - Diamond Light Source Ltd\I16\Nexus_Format\example_nexus\1041304.nxs")
    # root.mainloop()

    # root = create_root('Image')
    # fig, ax, ax_image, colorbar, toolbar = ini_image(root)
    # ax_image.remove()
    # ax_image = ax.pcolormesh(10 * np.random.rand(200, 50), shading='auto')
    # ax.set_xlim([0, 50])
    # ax.set_ylim([0, 200])
    # fig.canvas.draw()
    # root.mainloop()







"""
tk widget for viewing tree structure of HDF files
"""

import tkinter as tk
from tkinter import ttk

import h5py
import hdfmap

from mmg_toolbox.utils.file_functions import hdfobj_string
from ..misc.functions import open_close_all_tree, select_hdf_file
from ..misc.search import search_tree
from ..misc.styles import update_text_style
from .nexus_string_views import HdfNexusStr, HdfTreeStr, Nexus2SrsStr, NxTransformationsStr
from .nexus_treeview import HdfTreeview, HdfNameSpace


DETAILS_TAB_WIDTH = 30


class HDFViewer:
    """
    HDF Viewer - display cascading hierarchical data within HDF file in ttk GUI
        HDFViewer("filename.h5")
    Simple ttk interface for browsing HDF file structures.
     - Click Browse or File>Select File to pick an HDF, H5 or NeXus file
     - Collapse and expand the tree to view the file structure
     - Search for addresses using the search bar
     - Click on a dataset or group to view stored attributes and data

    :param root: tk root
    :param hdf_filename: str or None*, if str opens this file initially
    """

    def __init__(self, root: tk.Misc, hdf_filename: str = None, config: dict | None = None):
        self.map: hdfmap.NexusMap | None = None
        self.root = root
        self.config = config

        # Variables
        self.filepath = tk.StringVar(self.root, '')
        self.expandall = tk.BooleanVar(self.root, False)
        self.expression_box = tk.StringVar(self.root, '')
        self.expression_path = tk.StringVar(self.root, 'path = ')
        self.search_box = tk.StringVar(self.root, '')
        self.search_matchcase = tk.BooleanVar(self.root, False)
        self.search_wholeword = tk.BooleanVar(self.root, True)

        "------- Build Elements -----"
        # filepath
        self.ini_browse(self.root)

        main = ttk.Frame(self.root)
        main.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)

        frm = ttk.Frame(main)
        frm.pack(side=tk.LEFT, expand=tk.YES, fill=tk.BOTH)

        # Tabs
        self.view_tabs = ttk.Notebook(frm)
        tab1 = ttk.Frame(self.view_tabs)
        tab2 = ttk.Frame(self.view_tabs)
        tab3 = ttk.Frame(self.view_tabs)
        tab4 = ttk.Frame(self.view_tabs)
        tab5 = ttk.Frame(self.view_tabs)
        tab6 = ttk.Frame(self.view_tabs)
        self.view_tabs.bind('<<NotebookTabChanged>>', self.tab_change)

        self.view_tabs.add(tab1, text='HDF Tree')
        self.view_tabs.add(tab2, text='NeXus')
        self.view_tabs.add(tab3, text='HdfMap')
        self.view_tabs.add(tab4, text='Tree String')
        self.view_tabs.add(tab5, text='NeXus2SRS')
        self.view_tabs.add(tab6, text='NXtransformations')
        self.view_tabs.pack(side=tk.TOP, fill=tk.BOTH, expand=tk.YES)
        # treeviews
        self.hdf_tree = HdfTreeview(tab1)
        self.nexus = HdfNexusStr(tab2)
        self.hdf_map = HdfNameSpace(tab3)
        self.hdf_text = HdfTreeStr(tab4)
        self.nexus2srs = Nexus2SrsStr(tab5)
        self.transformations = NxTransformationsStr(tab6)

        self.hdf_tree.tree.bind('<<TreeviewSelect>>', self.tree_select)
        self.hdf_map.tree.bind('<<TreeviewSelect>>', self.tree_select)

        frm = ttk.Frame(main)
        frm.pack(side=tk.LEFT, expand=tk.NO, fill=tk.BOTH)
        # notebook
        tab_detail, tab_search, tab_expr = self.ini_notebook(frm)

        self.text = self.ini_details(tab_detail)
        self.ini_search(tab_search)
        self.text2 = self.ini_expression(tab_expr)

        if hasattr(self.root, 'style'):
            update_text_style(self.text, self.root.style)
            update_text_style(self.text2, self.root.style)

        "-------- Start Mainloop ------"
        if hdf_filename:
            self.filepath.set(hdf_filename)
            self.populate_from_file()

    "======================================================"
    "================= init functions ====================="
    "======================================================"

    def ini_browse(self, frame: tk.Misc):
        frm = ttk.Frame(frame)
        frm.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)

        var = ttk.Button(frm, text='Browse', command=self.select_file, width=10)
        var.pack(side=tk.LEFT)

        var = ttk.Entry(frm, textvariable=self.filepath)
        var.pack(side=tk.LEFT, expand=tk.YES, fill=tk.BOTH)
        var.bind('<Return>', self.populate_from_file)
        var.bind('<KP_Enter>', self.populate_from_file)

        var = ttk.Checkbutton(frm, variable=self.expandall, text='Expand', command=self.check_expand)
        var.pack(side=tk.LEFT)

    def ini_notebook(self, frame: tk.Misc) -> tuple[ttk.Frame, ttk.Frame, ttk.Frame]:

        frm = ttk.Frame(frame)
        frm.pack(side=tk.TOP, fill=tk.BOTH, expand=tk.YES)

        tab_control = ttk.Notebook(frm)
        tab1 = ttk.Frame(tab_control)
        tab2 = ttk.Frame(tab_control)
        tab3 = ttk.Frame(tab_control)

        tab_control.add(tab1, text='Details')
        tab_control.add(tab2, text='Search')
        tab_control.add(tab3, text='Expression')
        tab_control.pack(side=tk.TOP, fill=tk.BOTH, expand=tk.YES)

        return tab1, tab2, tab3

    def ini_details(self, frame: tk.Misc) -> tk.Text:
        frm = ttk.Frame(frame)
        frm.pack(side=tk.LEFT, expand=tk.YES, fill=tk.BOTH)

        text = tk.Text(frm, wrap=tk.NONE, width=DETAILS_TAB_WIDTH)
        text.pack(fill=tk.BOTH, expand=tk.YES)

        var = tk.Scrollbar(frm, orient=tk.HORIZONTAL, command=text.xview)
        var.pack(side=tk.BOTTOM, fill=tk.X)
        text.configure(xscrollcommand=var.set)
        return text

    def ini_search(self, frame: tk.Misc):
        frm = ttk.Frame(frame)
        frm.pack(side=tk.TOP, expand=tk.YES, fill=tk.X)

        var = ttk.Entry(frm, textvariable=self.search_box)
        var.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)
        # var.bind('<KeyRelease>', self.fun_search)
        var.bind('<Return>', self.fun_search)
        var.bind('<KP_Enter>', self.fun_search)
        var = ttk.Button(frm, text='Search', command=self.fun_search, width=10)
        var.pack(side=tk.TOP)

        line = ttk.Frame(frm)
        line.pack(side=tk.TOP)
        var = ttk.Checkbutton(line, variable=self.search_matchcase, text='Case')
        var.pack(side=tk.LEFT)
        var = ttk.Checkbutton(line, variable=self.search_wholeword, text='Word')
        var.pack(side=tk.LEFT)

    def ini_expression(self, frame: tk.Misc) -> tk.Text:
        frm = ttk.Frame(frame)
        frm.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)

        var = ttk.Entry(frm, textvariable=self.expression_box)
        var.pack(side=tk.TOP, fill=tk.X, expand=tk.YES)
        # var.bind('<KeyRelease>', self.fun_expression_reset)
        var.bind('<Return>', self.fun_expression)
        var.bind('<KP_Enter>', self.fun_expression)

        var = ttk.Label(frm, textvariable=self.expression_path)
        var.pack(side=tk.TOP, expand=tk.YES, fill=tk.X)

        var = ttk.Button(frm, text='Evaluate Expression', command=self.fun_expression)
        var.pack(side=tk.TOP, fill=tk.X, expand=tk.YES)

        text = tk.Text(frm, wrap=tk.NONE, width=DETAILS_TAB_WIDTH)
        text.pack(fill=tk.BOTH, expand=tk.YES)

        var = tk.Scrollbar(frm, orient=tk.HORIZONTAL, command=text.xview)
        var.pack(side=tk.BOTTOM, fill=tk.X)
        text.configure(xscrollcommand=var.set)
        return text

    "======================================================"
    "================ general functions ==================="
    "======================================================"

    def check_expand(self):
        selected = self.view_tabs.select()
        tab_name = self.view_tabs.tab(selected)['text']
        if tab_name == 'HDF Tree':
            open_close_all_tree(self.hdf_tree.tree, "", self.expandall.get())
        elif tab_name == 'HdfMap':
            open_close_all_tree(self.hdf_map.tree, "", self.expandall.get())

    def _delete_tree(self):
        self.hdf_tree.delete()
        self.hdf_map.delete()

    def populate_from_file(self, event=None):
        filename = self.filepath.get()
        self.map = hdfmap.create_nexus_map(filename)
        with hdfmap.load_hdf(filename) as hdf:
            self.populate(hdf, self.map)

    def populate(self, hdf_obj: h5py.File, hdf_map: hdfmap.NexusMap):
        self._delete_tree()
        self.hdf_tree.populate(hdf_obj, openstate=self.expandall.get())
        self.hdf_map.populate(hdf_obj, hdf_map)

    "======================================================"
    "================= event functions ===================="
    "======================================================"

    def tab_change(self, event):
        # Selected tab:
        selected = self.view_tabs.select()
        tab_name = self.view_tabs.tab(selected)['text']
        if tab_name == 'NeXus':
            self.nexus.populate(self.map)
        if tab_name == 'Tree String':
            self.hdf_text.populate(self.map.filename)
        if tab_name == 'NeXus2SRS':
            self.nexus2srs.populate(self.map)
        if tab_name == 'NXtransformations':
            self.transformations.populate(self.map.filename)

    def tree_select(self, event):
        self.text.delete('1.0', tk.END)
        tree = event.widget
        item = tree.focus()
        if 'path' in tree['columns']:  # hdfmap tab
            path = tree.set(item, column='path')
        else:
            path = tree.item(item, 'text')
        out = hdfobj_string(self.filepath.get(), path)
        self.text.insert('1.0', out)

    def select_file(self, event=None):
        filename = select_hdf_file(self.root)
        if filename:
            self.filepath.set(filename)
            self.populate_from_file()

    def fun_search(self, event=None):
        """Search currently active tab"""
        if self.view_tabs.index(self.view_tabs.select()) == 0:
            self.hdf_tree.tree.selection_remove(self.hdf_tree.tree.selection())
            search_tree(
                treeview=self.hdf_tree.tree,
                branch="",
                query=self.search_box.get(),
                match_case=self.search_matchcase.get(),
                whole_word=self.search_wholeword.get()
            )
        else:
            self.hdf_map.tree.selection_remove(self.hdf_map.tree.selection())
            search_tree(
                treeview=self.hdf_map.tree,
                branch="",
                query=self.search_box.get(),
                match_case=self.search_matchcase.get(),
                whole_word=self.search_wholeword.get()
            )

    def fun_expression(self, event=None):
        if self.map is None:
            return
        # self.text2.delete('1.0', tk.END)
        expression = self.expression_box.get()
        self.expression_path.set(f"path = {self.map.get_path(expression)}")
        out_str = f">>> {expression}\n"
        try:
            # out = hdfmap.hdf_eval(self.filepath.get(), expression)
            out = self.map.eval(self.map.load_hdf(), expression)
        except NameError as ne:
            out = ne
        out_str += f"{out}\n\n"
        # self.text2.insert('1.0', out_str)
        self.text2.insert(tk.END, out_str)




"""
{{title}}
{{description}}

Multi-peak fitting vs {{value}}
{{date}}
"""

import matplotlib.pyplot as plt
from mmg_toolbox import Experiment

data_dir = '{{experiment_dir}}'
scan_numbers = {{scan_numbers}}

exp = Experiment(data_dir, instrument='{{beamline}}')
exp.plot.set_plot_defaults()

scans = exp.scans(*scan_numbers)
# Fitting
amplitude = []
amplitude_err = []
metadata = []
for scan in scans:
    scan.fit.multi_peak_fit(
        xaxis='{{x-axis}}',
        yaxis='{{y-axis}}',
        npeaks=1,
        min_peak_power=None,
        peak_distance_idx=6,
        model='Gaussian',
        background='Slope'
    )
    print(scan.fit.fit_report())
    amp, err = scan.fit.fit_parameter('amplitude')
    amplitude.append(amp)
    amplitude_err.append(err)
    value, = scan.get_data('{{value}}', default=0)
    metadata.append(value)

fig, ax = plt.subplots()
ax.errorbar(metadata, amplitude, amplitude_err, fmt='.-', label='{{y-axis}}')
ax.set_xlabel('{{x-axis}}')
ax.set_ylabel('{{y-axis}}')
ax.set_title(exp.generate_scans_title(*scan_numbers))


plt.show()



"""
Useful tkinter functions that use matplotlib
"""
import matplotlib.pyplot as plt
import numpy as np
import tkinter as tk
from tkinter import ttk

from matplotlib.figure import Figure
from matplotlib.axes import Axes
from matplotlib.lines import Line2D
from matplotlib.patches import Rectangle
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk

from .styles import create_root, get_style_background

# parameters
FIGURE_SIZE = (8, 3)
IMAGE_SIZE = (8, 3)
FIGURE_DPI = 60
SMALL_FIGURE_DPI = 40
COLORMAPS = ['viridis', 'Spectral', 'plasma', 'inferno', 'Greys', 'Blues', 'winter', 'autumn',
             'hot', 'hot_r', 'hsv', 'rainbow', 'jet', 'twilight', 'hsv']
DEFAULT_COLORMAP = 'twilight'


class CustomToolbar(NavigationToolbar2Tk):
    """Customised version of matplotlib toolbar with added popout and copy functions"""

    def copy_button(self):
        """Copy figure to clipboard - doesn't currently work"""
        import io
        from PIL import Image
        # print(self.canvas.figure.canvas.tostring_rgb())
        image_buffer, (width, height) = self.canvas.figure.canvas.print_to_buffer()
        img = Image.frombytes("RGBA", (width, height), image_buffer)
        io_buffer = io.BytesIO()
        img.save(io_buffer, format='PNG')
        io_buffer.seek(0)
        self.master.clipboard_clear()
        self.master.clipboard_append(io_buffer.getvalue(), format="image/png")  # adds byte array to buffer but isn't interpreted

    def popout_figure(self):
        """Create a new tk window and display figure"""
        root = create_root('Figure', parent=self.master)
        fig, ax1, plot_list, toolbar = ini_plot(root, FIGURE_SIZE, FIGURE_DPI)

        for old_ax in self.canvas.figure.get_axes():
            ax1.set_title(old_ax.title.get_text())
            ax1.set_xlabel(old_ax.get_xlabel())
            ax1.set_ylabel(old_ax.get_ylabel())
            for line in old_ax.lines:
                ax1.plot(line.get_xdata(), line.get_ydata())

    def __init__(self, canvas_, parent_):
        # Add additional functions
        self.toolitems += (
            # (name, description, image, function)
            (None, None, None, None),  # seperator
            # ('copy', 'Copy Figure', 'filesave', 'copy_button'),
            ('popout', 'Popout Figure', 'qt4_editor_options', 'popout_figure'),
        )

        NavigationToolbar2Tk.__init__(self, canvas_, parent_)
        bg = get_style_background(parent_)
        self.config(background=bg)


def ini_plot(frame: tk.Misc, figure_size: tuple[int, int] | None = None,
             figure_dpi: int | None = None) -> tuple[Figure, Axes, list[Line2D], NavigationToolbar2Tk]:
    """Create a lineplot on a tk canvas with toolbar"""
    if figure_size is None:
        figure_size = FIGURE_SIZE
    if figure_dpi is None:
        figure_dpi = FIGURE_DPI

    # get the current background
    bg = get_style_background(frame)

    fig = Figure(figsize=figure_size, dpi=figure_dpi)
    try:
        fig.patch.set_facecolor(bg)
    except ValueError:
        print(f"Cannot set background color of {bg}")
        bg = '#dcdad5'
    # fig.subplots_adjust(left=0.2, bottom=0.2)
    # Amplitude
    ax1 = fig.add_subplot(111)
    ax1.set_autoscaley_on(True)
    ax1.set_autoscalex_on(True)
    ax1.set_xlabel(u'Axis 0')
    ax1.set_ylabel(u'Axis 1')
    ax1.set_title('filename')
    plot_list: list[plt.Line2D] = []

    frm = ttk.Frame(frame)
    # TODO: provide pack options input to control packing options
    frm.pack(side=tk.LEFT, expand=tk.YES, fill=tk.BOTH, pady=2, padx=5)
    canvas = FigureCanvasTkAgg(fig, frm)
    canvas.get_tk_widget().configure(bg='black')
    canvas.draw()
    canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=tk.YES, padx=5, pady=2)

    # Toolbar
    frm2 = ttk.Frame(frm)
    frm2.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH, padx=5, pady=2)
    # toolbar = NavigationToolbar2Tk(canvas, frm)
    toolbar = CustomToolbar(canvas, frm2)
    toolbar.config(background=bg)
    toolbar.update()
    toolbar.pack(fill=tk.X)#, expand=tk.YES)
    return fig, ax1, plot_list, toolbar


def ini_image(frame: tk.Misc, figure_size: tuple[int, int] | None = None, figure_dpi: int | None = None):
    """Create an image plot on a tk canvas with toolbar"""
    if figure_size is None:
        figure_size = IMAGE_SIZE
    if figure_dpi is None:
        figure_dpi = FIGURE_DPI

    # get the current background
    bg = get_style_background(frame)

    fig = Figure(figsize=figure_size, dpi=figure_dpi)
    try:
        fig.patch.set_facecolor(bg)
    except ValueError:
        print(f"Cannot set background color of {bg}")
        bg = '#dcdad5'

    ax1 = fig.add_subplot(111)
    # zeros = np.array([[0 for n in range(10)] for m in range(10)])
    xvals = np.arange(100)
    yvals = np.arange(100)
    default = np.random.rand(100, 100)
    ax1_image = ax1.pcolormesh(xvals, yvals, default, shading='auto', cmap=DEFAULT_COLORMAP)
    ax1.set_xlabel(u'Axis 0')
    ax1.set_ylabel(u'Axis 1')
    ax1.set_xlim([0, 100])
    ax1.set_ylim([0, 100])
    cb1 = fig.colorbar(ax1_image, ax=ax1)
    ax1.axis('image')
    plot_list: list[plt.Line2D] = []

    frm = ttk.Frame(frame)
    frm.pack(expand=tk.YES, fill=tk.BOTH, pady=2, padx=5)
    # frm.configure(bg=bg)
    canvas = FigureCanvasTkAgg(fig, frm)
    # canvas.get_tk_widget().configure(bg=bg)
    canvas.draw()
    canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=tk.YES, padx=5, pady=2)

    # Toolbar
    frm2 = ttk.Frame(frm)
    frm2.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH, padx=5, pady=2)
    # toolbar = NavigationToolbar2Tk(canvas, frm)
    toolbar = CustomToolbar(canvas, frm2)
    toolbar.config(background=bg)
    toolbar.update()
    toolbar.pack(fill=tk.X, expand=tk.YES)
    return fig, ax1, plot_list, ax1_image, cb1, toolbar


def add_rectangle(ax: Axes, left: float, bottom: float, width: float, height: float) -> Rectangle:
    """Add rectangle to axes"""
    rect = Rectangle((left, bottom), width, height, fill=False, edgecolor='black', facecolor='white', zorder=2)
    ax.add_patch(rect)
    return rect


"""
mmg_toolbox tests
Test nexus reader
"""

from mmg_toolbox.nexus.nexus_scan import NexusScan, NexusDataHolder
from mmg_toolbox.nexus.nexus_reader import (read_nexus_file, read_nexus_files, find_scans,
                                            find_similar_scans, find_matching_scans)
from . import only_dls_file_system
from .example_files import DIR

@only_dls_file_system
def test_read_nexus_file():
    f = DIR + '/i16/1109527.nxs'
    scan = read_nexus_file(f)
    assert isinstance(scan, NexusDataHolder)
    assert len(scan.eta_fly) == 61
    assert scan.roi2_sum.max() == 692919
    assert abs(scan.metadata.Tsample-300) < 0.1
    assert scan.scan_number() == 1109527
    assert scan.title() == '#1109527'


@only_dls_file_system
def test_read_nexus_files():
    files = [DIR + f'/i16/cm37262-1/{n}.nxs' for n in range(1032120, 1032130)]
    scans = read_nexus_files(*files)
    assert isinstance(scans[0], NexusScan)


@only_dls_file_system
def test_find_scans():
    files = [DIR + f'/i16/cm37262-1/{n}.nxs' for n in range(1032115, 1032135)]
    found_files = find_scans(*files, scan_command='pil3_100k')
    assert len(found_files) == 7
    found_files = find_scans(*files, scan_command='merlin', Ta=(300,0.2))
    assert len(found_files) == 2
    match  = {'start_time.timestamp()': (1705399908, 5)}
    found_files = find_scans(*files, **match)
    assert len(found_files) == 2




"""
Functions to load data from i06-1 and i10-1 beamline XAS measurements
"""

import numpy as np
import h5py
import hdfmap
import datetime

from mmg_toolbox.utils.file_functions import get_scan_number
from mmg_toolbox.utils.file_reader import read_dat_file
from mmg_toolbox.utils.polarisation import check_polarisation
from mmg_toolbox.nexus.nexus_functions import nx_find, nx_find_all, nx_find_data
from mmg_toolbox.beamline_metadata.hdfmap_generic import HdfMapXASMetadata as Md

from .spectra_analysis import energy_range_edge_label
from .spectra import Spectra
from .spectra_container import SpectraContainer, XasMetadata


def is_nxxas(filename: str) -> bool:
    """Return True if the NeXus file contains an entry or sub-entry with application definition NXxas"""
    return bool(nx_find_data(hdfmap.load_hdf(filename), 'NXentry', 'definition') == 'NXxas')


def create_xas_scan(name, energy: np.ndarray, monitor: np.ndarray, raw_signals: dict[str, np.ndarray],
                    filename: str = '', beamline: str = '', scan_no: int = 0, start_date_iso: str = '',
                    end_date_iso: str = '', cmd: str = '', default_mode: str = 'tey', pol: str = 'pc',
                    sample_name: str = '', temp: float = 300, mag_field: float = 0, pitch: float = 0,
                    element_edge: str | None = None) -> SpectraContainer:
    """
    Function to load data from i06-1 and i10-1 beamline XAS measurements
    """
    # Check spectra
    if default_mode not in raw_signals:
        raise KeyError(f"mode '{default_mode}' is not available in {list(raw_signals.keys())}")
    if element_edge is None:
        element, edge = energy_range_edge_label(energy.min(), energy.max())
    else:
        element, edge = element_edge.split()

    for detector, array in raw_signals.items():
        if len(array) != len(energy):
            print(f"Removing signal '{detector}' as the length is wrong")
        if np.max(array) < 0.1:
            print(f"Removing signal '{detector}' as the values are 0")
    raw_signals = {
        detector: array for detector, array in raw_signals.items()
        if len(array) == len(energy) and array.max() > 0.1
    }

    if len(raw_signals) == 0:
        raise ValueError("No raw_signals found")

    # perform Analysis steps
    spectra = {
        name: Spectra(energy, signal / monitor, mode=name, process_label='raw', process=f"{name} / monitor")
        for name, signal in raw_signals.items()
    }
    metadata = XasMetadata(
        filename=filename,
        beamline=beamline,
        scan_no=scan_no,
        start_date_iso=start_date_iso,
        end_date_iso=end_date_iso,
        cmd=cmd,
        default_mode=default_mode,
        pol=pol,
        sample_name=sample_name,
        temp=temp,
        mag_field=mag_field,
        pitch=pitch,
        element=element,
        edge=edge,
        energy=energy,
        raw_signals=raw_signals,
        monitor=monitor
    )
    return SpectraContainer(name, spectra, metadata=metadata)


def load_from_dat(filename: str, sample_name='', element_edge=None) -> SpectraContainer:
    # read file
    scan = read_dat_file(filename)

    # read Scan data
    scannables = scan.keys()
    mode = 'tey'
    energy = next(iter(scan.values()))  # fastEnergy or something else
    if 'C1' in scannables:
        signals = {
            'tey': scan['C1'],
            'tfy': scan['C3'],
        }
        monitor = scan['C2']
    elif 'msc19' in scannables:
        signals = {
            'tey': scan['msc19'],
            'tfy': scan['msc18'],
        }
        monitor = scan['msc19']
    else:
        raise ValueError(f"file {filename} does not contain a known signal scannable")

    # read Metadata
    metadata = scan.metadata
    beamline = metadata.get('SRSPRJ', '?').lower()
    pol_from_id = 'pc' if metadata.get('iddtrp', 1) > 0 else 'nc'
    pol = check_polarisation(metadata.get('polarisation', pol_from_id))
    temp = metadata.get('Tsample_mag', 300)
    mag_field = metadata.get('magz', 0)
    cmd = metadata.get('command', '')
    date_str = metadata.get('date', '')
    date = datetime.datetime.strptime(date_str, '%a %b %d %H:%M:%S %Y')
    date_iso = date.isoformat()
    scan_no = metadata.get('SRSRUN', get_scan_number(filename))

    return create_xas_scan(
        name=str(scan_no),
        energy=energy,
        raw_signals=signals,
        monitor=monitor,
        filename=filename,
        beamline=beamline,
        scan_no=scan_no,
        start_date_iso=date_iso,
        end_date_iso=date_iso,
        cmd=cmd,
        default_mode=mode,
        pol=pol,
        sample_name=sample_name,
        temp=temp,
        mag_field=mag_field,
        element_edge=element_edge
    )


def load_from_nxs(filename: str, sample_name=None, element_edge=None) -> SpectraContainer:
    # read file
    with h5py.File(filename, 'r') as hdf:
        # read Scan data
        mode = nx_find(hdf, 'NXxas', 'NXdata', 'mode')
        if mode:
            # NXxas application definition
            energy = nx_find_data(hdf, 'NXxas', 'NXdata', 'axes')
            monitor = nx_find_data(hdf, 'NXxas', 'NXmonitor', 'signal')
            signals = {
                nx_find_data(grp, 'mode'): nx_find_data(grp, 'signal')
                for grp in nx_find_all(hdf, 'NXxas', 'NXdata')
            }
        else:
            mode = 'tey'
            energy = nx_find_data(hdf, 'NXdata', 'axes')
            if '/entry/instrument/fesData/C1' in hdf:
                # i06-1 old nexus
                monitor = hdf['/entry/instrument/fesData/C2'][()]
                signals = {
                    'tey': hdf['/entry/instrument/fesData/C1'][()],
                    'tfy': hdf['/entry/instrument/fesData/C3'][()],
                }
                if signals['tfy'].max() < 0.1:
                    signals['tfy'] = hdf['/entry/instrument/fesData/C4'][()]
            elif '/entry/mcse16/data' in hdf:
                # i10-1 old nexus
                monitor = hdf['/entry/mcse16/data'][()]
                signals = {
                    'tey': hdf['/entry/mcse17/data'][()],
                    'tfy': hdf['/entry/mcse19/data'][()],
                }
            else:
                nxdata = nx_find(hdf, 'NXdata')
                fields = list(nxdata.keys())
                raise ValueError(f'Unknown data fields: {fields}')

        # read Metadata - currently only works for i06-1!
        beamline = nx_find_data(hdf, 'NXinstrument', 'name', default='?')
        pol = nx_find_data(hdf, 'NXinsertion_device', 'polarisation', default='?')
        temp = nx_find_data(hdf, '/entry/instrument/scm/T_sample', default=300)
        mag_field = nx_find_data(hdf, '/entry/instrument/scm/field_z', default=0)
        cmd = nx_find_data(hdf, 'scan_command', default='')
        start_date_iso = nx_find_data(hdf, 'start_time', default='')
        end_date_iso = nx_find_data(hdf, 'end_time', default=start_date_iso)
        scan_no = nx_find_data(hdf, 'entry_identifier', default=get_scan_number(filename))

        if sample_name is None:
            sample_name = nx_find_data(hdf, 'NXsample', 'name', default='')

    return create_xas_scan(
        name=str(scan_no),
        energy=energy,
        raw_signals=signals,
        monitor=monitor,
        filename=filename,
        beamline=beamline,
        scan_no=scan_no,
        start_date_iso=start_date_iso,
        end_date_iso=end_date_iso,
        cmd=cmd,
        default_mode=mode,
        pol=pol,
        sample_name=sample_name,
        temp=temp,
        mag_field=mag_field,
        element_edge=element_edge
    )


def load_from_nxs_using_hdfmap(filename: str, sample_name: str | None = None,
                               element_edge: str | None = None) -> SpectraContainer:
    """Load ScanContainer"""

    with hdfmap.load_hdf(filename) as hdf:
        # HdfMap creates data-path namespace
        m = hdfmap.NexusMap()
        m.populate(hdf)

        # scan data
        scan_no = m.eval(hdf, 'entry_identifier', default=get_scan_number(filename))
        energy = m.eval(hdf, Md.energy)
        monitor = m.eval(hdf, Md.monitor)
        mode = 'tey'
        signals = {
            'tey': m.eval(hdf, Md.tey),
            'tfy': m.eval(hdf, Md.tfy),
        }
        # metadata
        beamline = m.eval(hdf, 'f"{beamline}_{end_station}" if end_station else instrument_name')
        start_date_iso = m.eval(hdf, 'str(start_time)')
        end_date_iso = m.eval(hdf, 'str(end_time)')
        cmd = m.eval(hdf, Md.cmd)
        pol = check_polarisation(m.eval(hdf, Md.pol))
        if sample_name is None:
            sample_name = m.eval(hdf, 'sample_name', '')
        temp = m.eval(hdf, Md.temp)
        mag_field = m.eval(hdf, Md.field_z)
        pitch = m.eval(hdf, Md.rot)
    return create_xas_scan(
        name=str(scan_no),
        energy=energy,
        raw_signals=signals,
        monitor=monitor,
        filename=filename,
        beamline=beamline,
        scan_no=scan_no,
        start_date_iso=start_date_iso,
        end_date_iso=end_date_iso,
        cmd=cmd,
        default_mode=mode,
        pol=pol,
        sample_name=sample_name,
        temp=temp,
        mag_field=mag_field,
        pitch=pitch,
        element_edge=element_edge
    )


def load_xas_scans(*filenames: str, sample_name: str | None = None, element_edge: str | None = None) -> list[SpectraContainer]:
    """Load scans from a list of filenames, return {'pol': [scan1, scan2, ...]}"""
    scans = [
        load_from_dat(filename, sample_name=sample_name, element_edge=element_edge)
        if filename.endswith('.dat')
        else load_from_nxs_using_hdfmap(filename, sample_name=sample_name, element_edge=element_edge)
        for filename in filenames
    ]
    return scans


def find_similar_measurements(*filenames: str, temp_tol=1., field_tol=0.1) -> list[SpectraContainer]:
    """
    Find similar measurements based on energy, temperature and field.

    Each measurement is compared to the first one in the list, using energy, temperature and field tolerances.

    The polarisation is also checked to be similar (lh, lv or cl, cr).

    Scans with different or missing metadata are removed from the list.

    :param filenames: List of filenames to compare
    :param temp_tol: Tolerance for temperature comparison (default: 0.1 K)
    :param field_tol: Tolerance for field comparison (default: 0.1 T)
    :return: List of similar measurements
    """
    from mmg_toolbox.nexus.nexus_reader import find_matching_scans
    ini_scan = load_xas_scans(filenames[0])[0]
    if len(filenames) == 1:
        filenames = find_matching_scans(filenames[0])
    element = ini_scan.metadata.element
    edge = ini_scan.metadata.edge
    temperature = ini_scan.metadata.temp
    field_z = abs(ini_scan.metadata.mag_field)  # allow +/- field
    pol = ini_scan.metadata.pol
    if pol in ['lh', 'lv']:
        similar_pols = ['lh', 'lv']
    elif pol in ['cl', 'cr']:
        similar_pols = ['cl', 'cr']
    elif pol in ['nc', 'pc']:
        similar_pols = ['nc', 'pc']
    else:
        raise ValueError(f"Unknown polarisation: {pol}")

    similar = []
    for filename in filenames:
        try:
            scan = load_xas_scans(filename)[0]
        except ValueError as ve:
            print(f"Error loading {filename} as xas_scan: {ve}")
            continue
        m = scan.metadata
        if (
            m.element == element and
            m.edge == edge and
            abs(m.temp - temperature) < temp_tol and
            abs(abs(m.mag_field) - field_z) < field_tol and
            m.pol in similar_pols
        ):
            similar.append(scan)
        else:
            print(f"Measurement {repr(scan)} is not similar to {repr(ini_scan)}")
    return similar




"""
X-ray scattering utility funcitons
"""

import numpy as np


# Constants
class Const:
    pi = np.pi  # mmmm tasty Pi
    e = 1.6021733E-19  # C  electron charge
    h = 6.62606868E-34  # Js  Plank consant
    c = 299792458  # m/s   Speed of light
    u0 = 4 * pi * 1e-7  # H m-1 Magnetic permeability of free space
    me = 9.109e-31  # kg Electron rest mass
    mn = 1.6749e-27 # kg Neutron rest mass
    Na = 6.022e23  # Avagadro's No
    A = 1e-10  # m Angstrom
    r0 = 2.8179403227e-15  # m classical electron radius = e^2/(4pi*e0*me*c^2)
    Cu = 8.048  # Cu-Ka emission energy, keV
    Mo = 17.4808  # Mo-Ka emission energy, keV


def photon_wavelength(energy_kev):
    """
    Converts energy in keV to wavelength in A
     wavelength_a = photon_wavelength(energy_kev)
     lambda [A] = h*c/E = 12.3984 / E [keV]
    """

    # Electron Volts:
    E = 1000 * energy_kev * Const.e

    # SI: E = hc/lambda
    lam = Const.h * Const.c / E
    wavelength = lam / Const.A
    return wavelength


def photon_energy(wavelength_a):
    """
    Converts wavelength in A to energy in keV
     energy_kev = photon_energy(wavelength_a)
     Energy [keV] = h*c/L = 12.3984 / lambda [A]
    """

    # SI: E = hc/lambda
    lam = wavelength_a * Const.A
    E = Const.h * Const.c / lam

    # Electron Volts:
    energy = E / Const.e
    return energy / 1000.0


" --------------------------------------------------------------------- "
" --------------------------- Wavevectors ----------------------------- "
" --------------------------------------------------------------------- "


def wavevector(wavelength_a):
    """Return wavevector = 2pi/lambda"""
    return 2 * Const.pi / wavelength_a


def resolution2energy(res, twotheta=180.):
    """
    Calcualte the energy required to achieve a specific resolution at a given two-theta
    :param res: measurement resolution in A (==d-spacing)
    :param twotheta: Bragg angle in Degrees
    :return: float
    """
    theta = twotheta * Const.pi / 360  # theta in radians
    return (Const.h * Const.c * 1e10) / (res * np.sin(theta) * Const.e * 2 * 1000.)


def diffractometer_twotheta(delta=0, gamma=0):
    """Return the Bragg 2-theta angle for diffractometer detector rotations delta (vertical) and gamma (horizontal)"""
    delta = np.deg2rad(delta)
    gamma = np.deg2rad(gamma)
    twotheta = np.arccos(np.cos(delta) * np.cos(gamma))
    return np.rad2deg(twotheta)


def you_normal_vector(eta=0, chi=90, mu=0):
    """
    Determine the normal vector using the You diffractometer angles
      you_normal_vector(0, 0, 0) = [1, 0, 0]
      you_normal_vector(0, 90, 0) = [0, 1, 0]
      you_normal_vector(90, 90, 0) = [0, 0, -1]
      you_normal_vector(0, 0, 90) = [0, 0, -1]
    :param eta: angle (deg) along the x-axis
    :param mu: angle (deg) about the z-axis
    :param chi: angle deg) a
    :return: array
    """
    eta = np.deg2rad(eta)
    chi = np.deg2rad(chi)
    mu = np.deg2rad(mu)
    normal = np.array([np.sin(mu) * np.sin(eta) * np.sin(chi) + np.cos(mu) * np.cos(chi),
                       np.cos(eta) * np.sin(chi),
                       -np.cos(mu) * np.sin(eta) * np.sin(chi) - np.sin(mu) * np.cos(chi)])
    return normal


def wavevector_i(wavelength_a):
    """
    Returns a 3D wavevector for the initial wavevector
    """
    k = wavevector(wavelength_a)
    return np.array([0, 0, k])


def wavevector_f(wavelength_a, delta=0, gamma=0):
    """
    Returns a 3D wavevector for the final wavevector
    """
    k = wavevector(wavelength_a)
    sd = np.sin(np.deg2rad(delta))
    sg = np.sin(np.deg2rad(gamma))
    cd = np.cos(np.deg2rad(delta))
    cg = np.cos(np.deg2rad(gamma))
    return k * np.array([sg * cd, sd, cg * cd])


def wavevector_t(wavelength_a, delta=0, gamma=0):
    """
    Returns the wavevector transfer in inverse-Angstroms
      Q = kf - ki
    """
    ki = wavevector_i(wavelength_a)
    kf = wavevector_f(wavelength_a, delta, gamma)
    return kf - ki


def polarisation_sigma(delta=0, gamma=0):
    """
    Returns the scattered polerisation vector in the sigma' channel
    """
    sd = np.sin(np.deg2rad(delta))
    sg = np.sin(np.deg2rad(gamma))
    cd = np.cos(np.deg2rad(delta))
    cg = np.cos(np.deg2rad(gamma))
    return np.array([cg, 0, -sg])


def polarisation_pi(delta=0, gamma=0):
    """
    Returns the scattered polerisation vector in the Pi' channel
    """
    sd = np.sin(np.deg2rad(delta))
    sg = np.sin(np.deg2rad(gamma))
    cd = np.cos(np.deg2rad(delta))
    cg = np.cos(np.deg2rad(gamma))
    return np.array([-sg * sd, cd, -cg * sd])





"""
widget for running scripts
"""

import tkinter as tk
from tkinter import ttk, messagebox
import matplotlib.pyplot as plt

import hdfmap

from mmg_toolbox import Experiment
from mmg_toolbox.scripts import NOTEBOOKS, SCRIPTS, R
from mmg_toolbox.utils.env_functions import get_first_file, get_processing_directory
from ..misc.logging import create_logger
from ..misc.config import get_config, C
from ..misc.functions import select_folder
from ..misc.processing_options import create_script_from_template, create_notebook_from_template
from ..widgets.scan_range_selector import ScanRangeSelector

logger = create_logger(__file__)


class MultiScanAnalysis:
    """Frame with scan number generation and buttons for plotting and processing options"""

    def __init__(self, root: tk.Misc, config: dict | None = None, exp_directory: str | None = None,
                 proc_directory: str | None = None, scan_numbers: list[int] | None = None,
                 metadata: str | None = None, x_axis: str | None = None, y_axis: str | None = None):
        logger.info('Creating ScriptRunner')
        self.root = root
        self.config = config or get_config()

        if exp_directory is None:
            exp_directory = self.config.get(C.current_dir, '')
        if proc_directory is None:
            proc_directory = self.config.get(C.current_proc, get_processing_directory(exp_directory))

        self.exp_folder = tk.StringVar(root, exp_directory)
        self.proc_folder = tk.StringVar(root, proc_directory)
        self.output_file = tk.StringVar(root, proc_directory + '/file.py')
        self.x_axis = tk.StringVar(self.root, 'axes' if x_axis is None else x_axis)
        self.y_axis = tk.StringVar(self.root, 'signal' if y_axis is None else y_axis)
        self.metadata_name = tk.StringVar(self.root, metadata or config.get(C.default_metadata, ''))
        self.script_name = tk.StringVar(root, 'example')
        self.script_desc = tk.StringVar(root, '')
        self.options = {}
        self.file_list = []

        # sec = ttk.LabelFrame(self.root, text='Folders')
        # sec.pack(side=tk.TOP, fill=tk.BOTH, expand=tk.YES, padx=4, pady=4)
        #
        # frm = ttk.Frame(sec)
        # frm.pack(side=tk.TOP, fill=tk.X, expand=tk.YES, padx=4)
        # ttk.Label(frm, text='Data Dir:', width=15).pack(side=tk.LEFT, padx=4)
        # ttk.Entry(frm, textvariable=self.exp_folder, width=60).pack(side=tk.LEFT)
        # ttk.Button(frm, text='Browse', command=self.browse_datadir).pack(side=tk.LEFT)
        #
        # frm = ttk.Frame(sec)
        # frm.pack(side=tk.TOP, fill=tk.X, expand=tk.YES, padx=4)
        # ttk.Label(frm, text='Analysis Dir:', width=15).pack(side=tk.LEFT, padx=4)
        # ttk.Entry(frm, textvariable=self.proc_folder, width=60).pack(side=tk.LEFT)
        # ttk.Button(frm, text='Browse', command=self.browse_analysis).pack(side=tk.LEFT)

        # Axis + Metadata selection
        sec = ttk.LabelFrame(self.root, text='Axes')
        sec.pack(side=tk.TOP, fill=tk.BOTH, expand=tk.YES, padx=4, pady=4)

        frm = ttk.Frame(sec)
        frm.pack(side=tk.TOP, fill=tk.X, expand=tk.YES, padx=4)
        ttk.Label(frm, text='X:', width=2).pack(side=tk.LEFT, padx=2)
        ttk.Entry(frm, textvariable=self.x_axis, width=20).pack(side=tk.LEFT)
        ttk.Button(frm, text=':', command=self.browse_x_axis, width=1).pack(side=tk.LEFT)

        ttk.Label(frm, text='Y:', width=2).pack(side=tk.LEFT, padx=4)
        ttk.Entry(frm, textvariable=self.y_axis, width=20).pack(side=tk.LEFT)
        ttk.Button(frm, text=':', command=self.browse_y_axis, width=1).pack(side=tk.LEFT)

        ttk.Label(frm, text='Metadata:', width=15).pack(side=tk.LEFT, padx=4)
        ttk.Entry(frm, textvariable=self.metadata_name, width=20).pack(side=tk.LEFT)
        ttk.Button(frm, text=':', command=self.browse_metadata, width=1).pack(side=tk.LEFT)

        # Range selection
        sec = ttk.LabelFrame(self.root, text='Scan Numbers')
        sec.pack(side=tk.TOP, fill=tk.BOTH, expand=tk.YES, padx=4, pady=4)

        self.range = ScanRangeSelector(sec, exp_directory, self.config, metadata_getter=self.metadata_name)
        self.range.exp_folder.set(exp_directory)
        if scan_numbers is not None:
            self.range.text.insert("1.0", str(scan_numbers))

        line = ttk.Frame(self.root)
        line.pack(side=tk.TOP, expand=tk.YES, pady=8, padx=4)
        ttk.Button(line, text='Plot', command=self.plot_legend, width=10).pack(side=tk.LEFT)
        ttk.Button(line, text='Plot lines', command=self.plot_lines, width=10).pack(side=tk.LEFT)
        ttk.Button(line, text='Plot Meta', command=self.plot_metadata, width=10).pack(side=tk.LEFT)
        ttk.Button(line, text='Multi-Plot', command=self.multiplot, width=10).pack(side=tk.LEFT)
        ttk.Button(line, text='Plot 2D', command=self.plot2d, width=10).pack(side=tk.LEFT)
        ttk.Button(line, text='Plot 3D', command=self.plot3d, width=10).pack(side=tk.LEFT)
        ttk.Button(line, text='Plot Surf', command=self.plot3d, width=10).pack(side=tk.LEFT)

        line = ttk.Frame(self.root)
        line.pack(side=tk.TOP, expand=tk.YES, pady=8, padx=4)
        ttk.Button(line, text='Fits', command=self.fitting, width=10).pack(side=tk.LEFT)
        ttk.Button(line, text='Convert to dat', command=self.convert2dat).pack(side=tk.LEFT)

        # Scripts
        script_names = ('Scripts:',) + tuple(SCRIPTS) + ('Notebooks:',) + tuple(NOTEBOOKS)
        self.scripts = {f"{n:2} {name}": name for n, name in enumerate(script_names)}
        ttk.Button(line, text='Create', command=self.script_create).pack(side=tk.LEFT, padx=(20, 0))
        var = ttk.OptionMenu(line, self.script_name, 'Script', *self.scripts,
                             command=self.script_select)
        var.pack(side=tk.LEFT, padx=4)
        ttk.Label(line, textvariable=self.script_desc).pack(side=tk.LEFT)

    def browse_datadir(self):
        folder = select_folder(self.root)
        if folder:
            self.exp_folder.set(folder)

    def browse_analysis(self):
        folder = select_folder(self.root)
        if folder:
            self.proc_folder.set(folder)

    def browse_x_axis(self):
        from ..apps.namespace_select import create_scannable_selector
        scan_file = next(iter(self.range.generate_scan_files().values()), get_first_file(self.exp_folder.get()))
        hdf_map = hdfmap.create_nexus_map(scan_file)
        names = create_scannable_selector(hdf_map)
        if names:
            self.x_axis.set(', '.join(name for name in names))

    def browse_y_axis(self):
        from ..apps.namespace_select import create_scannable_selector
        scan_file = next(iter(self.range.generate_scan_files().values()), get_first_file(self.exp_folder.get()))
        hdf_map = hdfmap.create_nexus_map(scan_file)
        names = create_scannable_selector(hdf_map)
        if names:
            self.y_axis.set(', '.join(name for name in names))

    def browse_metadata(self):
        from ..apps.namespace_select import create_metadata_selector
        scan_file = next(iter(self.range.generate_scan_files().values()), get_first_file(self.exp_folder.get()))
        hdf_map = hdfmap.create_nexus_map(scan_file)
        paths = create_metadata_selector(hdf_map)
        if paths:
            self.metadata_name.set(', '.join(path for path in paths))

    def get_experiment(self):
        return Experiment(self.exp_folder.get(), instrument=self.config.get('beamline', None))

    def fitting(self):
        from ..apps.peak_fit_analysis import create_peak_fit
        scan_numbers = self.range.generate_scan_numbers()
        create_peak_fit(
            parent=self.root,
            config=self.config,
            exp_directory=self.exp_folder.get(),
            proc_directory=self.proc_folder.get(),
            scan_numbers=scan_numbers,
            metadata=self.metadata_name.get(),
            x_axis=self.x_axis.get(),
            y_axis=self.y_axis.get()
        )

    def convert2dat(self):
        from nexus2srs import run_nexus2srs
        scan_files = self.range.generate_scan_files()
        answer = messagebox.askokcancel(
            title='Nexus2SRS',
            message=f'Convert {len(scan_files)} NeXus files to .dat format?',
            icon='warning',
            parent=self.root
        )
        if answer:
            run_nexus2srs('-tiff', *scan_files.values())
            messagebox.showinfo(
                title='Nexus2SRS',
                message='Conversion complete!',
                parent=self.root
            )

    def plot_legend(self):
        exp = self.get_experiment()
        scan_numbers = self.range.generate_scan_numbers()
        xaxis = self.x_axis.get()
        yaxis = self.y_axis.get()
        exp.plot.plot(*scan_numbers, xaxis=xaxis, yaxis=yaxis)
        plt.show()

    def plot_lines(self):
        exp = self.get_experiment()
        scan_numbers = self.range.generate_scan_numbers()
        xaxis = self.x_axis.get()
        yaxis = self.y_axis.get()
        values = self.metadata_name.get()
        exp.plot.multi_lines(*scan_numbers, xaxis=xaxis, yaxis=yaxis, value=values)
        plt.show()

    def plot2d(self):
        exp = self.get_experiment()
        scan_numbers = self.range.generate_scan_numbers()
        xaxis = self.x_axis.get()
        yaxis = self.y_axis.get()
        values = self.metadata_name.get()
        values = values if values else None
        exp.plot.surface_2d(*scan_numbers, xaxis=xaxis, signal=yaxis, values=values)
        plt.show()

    def plot3d(self):
        exp = self.get_experiment()
        scan_numbers = self.range.generate_scan_numbers()
        xaxis = self.x_axis.get()
        yaxis = self.y_axis.get()
        values = self.metadata_name.get()
        values = values if values else None
        exp.plot.lines_3d(*scan_numbers, xaxis=xaxis, signal=yaxis, values=values)
        plt.show()

    def plot_surf(self):
        exp = self.get_experiment()
        scan_numbers = self.range.generate_scan_numbers()
        xaxis = self.x_axis.get()
        yaxis = self.y_axis.get()
        values = self.metadata_name.get()
        values = values if values else None
        exp.plot.surface_3d(*scan_numbers, xaxis=xaxis, signal=yaxis, values=values)
        plt.show()

    def multiplot(self):
        exp = self.get_experiment()
        scan_numbers = self.range.generate_scan_numbers()
        xaxis = self.x_axis.get()
        yaxis = self.y_axis.get()
        values = self.metadata_name.get()
        values = values if values else None
        exp.plot.multi_plot(*scan_numbers, xaxis=xaxis, yaxis=yaxis, value=values)
        plt.show()

    def plot_metadata(self):
        exp = self.get_experiment()
        scan_numbers = self.range.generate_scan_numbers()
        values = self.metadata_name.get().split(',')
        exp.plot.metadata(*scan_numbers, values=values)
        plt.show()

    def script_select(self, _event=None):
        script_label = self.script_name.get()
        script_index = int(script_label.split()[0])
        notebook_index = next(n for n, lab in enumerate(self.scripts.values()) if 'Notebooks:' in lab)
        script = self.scripts[script_label]
        if script_index < notebook_index and script in SCRIPTS:
            _, desc = SCRIPTS[script]
            desc = 'Script: ' + desc
            self.script_desc.set(desc)
        elif script in NOTEBOOKS:
            _, desc = NOTEBOOKS[script]
            desc = 'Notebook: ' + desc
            self.script_desc.set(desc)
        else:
            self.script_desc.set('')

    def script_create(self, _event=None):
        proc = self.proc_folder.get()
        script_name = self.script_name.get()
        if script_name not in self.scripts:
            return
        script = self.scripts[self.script_name.get()]
        exp = self.get_experiment()
        scan_numbers = self.range.generate_scan_numbers()
        if not scan_numbers:
            return
        scan_files = self.range.generate_scan_files()
        values = self.metadata_name.get().split(',')
        value = values[0] if values else None
        first_scan, = exp.scans(scan_numbers[0])

        replacements = {
            R.exp: self.exp_folder.get(),
            R.filepaths: '\n'.join(f"    '{f}'," for f in scan_files.values()),
            R.beamline: self.config.get(C.beamline, ''),
            R.scannos: str(scan_numbers),
            R.title: exp.generate_scans_title(*scan_numbers, metadata_str=value, hdf_map=first_scan.map),
            R.description: 'Created by  MultiScanAnalysis',
            R.xaxis: self.x_axis.get(),
            R.yaxis: self.y_axis.get(),
            R.value: self.metadata_name.get(),
        }

        if script in SCRIPTS:
            create_script_from_template(self.root, script, proc, self.config, **replacements)
        elif script in NOTEBOOKS:
            create_notebook_from_template(self.root, script, proc, self.config, **replacements)




"""
Jupyter Notebook commands
"""

from urllib.request import urlopen
import webbrowser
import subprocess

SUBPROCESSES = []


def popen_jupyter_server(cmd='notebook', directory: str | None = None, file: str | None = None):
    """
    Start a jupyter server quietly in the background.

    :param cmd: 'notebook' | 'lab'
    :param directory: None, or directory to start in
    :param file: None, or file to open
    """
    command = ['jupyter', cmd]
    if directory:
        command.append('--ServerApp.root_dir=' + directory)
    elif file:
        command.append(file)
    print(f"Running: {' '.join(command)}")
    output = subprocess.Popen(command, stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)
    SUBPROCESSES.append(output)


def check_notebook_servers() -> list[str]:
    list_servers = subprocess.run("jupyter server list", shell=True, capture_output=True)
    output = list_servers.stdout.decode()
    urls = [item for item in output.split() if item.startswith('http')]
    # Check urls exist
    for url in urls:
        try:
            with urlopen(url) as response:
                pass
        except OSError:
            urls.remove(url)
    print('Notebook Server URLs:')
    print(urls)
    return urls


def terminate_notebooks():
    subprocess.run("jupyter server stop", shell=True)


def launch_jupyter_notebook(cmd='notebook', directory: str | None = None, file: str | None = None):
    """
    Open a running jupyter server or launch one.

    :param cmd: 'notebook' | 'lab'
    :param directory: None, or directory to start in
    :param file: None, or file to open
    """
    running_notebooks = check_notebook_servers()
    if running_notebooks:
        webbrowser.open(running_notebooks[-1])
    else:
        popen_jupyter_server(cmd, directory, file)




"""
Simple window with editable text box
"""

from mmg_toolbox.tkguis.misc.styles import tk, ttk, create_root
from mmg_toolbox.tkguis.misc.logging import create_logger

logger = create_logger(__file__)


class EditText:
    """
    Simple text edit box
        next_text = EditText(old_text, tkframe).show()
    :param expression: str expression to edit
    :param parent: tk root
    """

    def __init__(self, expression: str, parent: tk.Misc | None = None, textwidth=30, title="Edit text"):
        self.output = expression
        self.root = create_root(title, parent=parent)

        frm = ttk.Frame(self.root)
        frm.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)

        self.text = tk.Text(frm, wrap=tk.NONE, width=textwidth)
        self.text.pack(side=tk.LEFT, fill=tk.BOTH, expand=tk.YES)
        self.text.insert('1.0', expression)

        var = ttk.Scrollbar(frm, orient=tk.VERTICAL, command=self.text.yview)
        var.pack(side=tk.LEFT, fill=tk.Y)
        self.text.configure(yscrollcommand=var.set)

        frm = ttk.Frame(self.root)
        frm.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)

        var = ttk.Button(self.root, text='Update', command=self.fun_update)
        var.pack(side=tk.TOP, fill=tk.X)

    def fun_update(self, event=None):
        """Launches window, returns selection"""
        self.output = self.text.get('1.0', tk.END)
        self.root.destroy()  # trigger wait_window

    def show(self):
        """Launches window, returns selection"""
        self.root.wait_window()  # wait for window
        self.root.destroy()
        return self.output


"""
a tkinter frame with a single plot
"""
import os
import tkinter as tk
from tkinter import ttk
from tkinter.messagebox import askyesnocancel

import hdfmap
from hdfmap import create_nexus_map

from mmg_toolbox.utils.env_functions import get_scan_notebooks, TMPDIR
from ..misc.functions import post_right_click_menu, show_error
from ..misc.logging import create_logger
from ..misc.config import get_config, C
from ..misc.screen_size import get_text_size
from ..apps.edit_text import EditText

logger = create_logger(__file__)


class NexusDetails:
    def __init__(self, root: tk.Misc, hdf_filename: str | None = None,
                 config: dict | None = None):
        self.root = root
        self.filename = hdf_filename
        self.map: hdfmap.NexusMap | None = None
        self.config = config or get_config()

        self.terminal_history = ['']
        self.terminal_history_index = 0
        self._text_expression = self.config.get(C.metadata_string, '')
        self.terminal_entry = tk.StringVar(self.root, self.terminal_history[self.terminal_history_index])
        self.notebook = tk.StringVar(self.root, 'None')
        self.notebooks = {}  # notebook: filepath

        self.terminal = self.ini_terminal(self.root)  # pack from bottom
        self.combo_notebook = self.ini_notebooks(self.root)  # pack from bottom
        self.textbox = self.ini_textbox(self.root)

        if hdf_filename:
            self.update_data_from_file(hdf_filename)

    def ini_textbox(self, frame: tk.Misc):
        frm = ttk.Frame(frame)
        frm.pack(side=tk.TOP, fill=tk.BOTH, expand=tk.YES)

        xfrm = ttk.Frame(frm)
        xfrm.pack(side=tk.TOP, expand=tk.YES, fill=tk.BOTH)

        # reduce text height if screen is small
        text_chars, text_lines = get_text_size(self.root, self.config)
        text = tk.Text(xfrm, state=tk.DISABLED, wrap=tk.NONE, width=text_chars, height=text_lines)
        text.pack(fill=tk.BOTH, expand=tk.YES)
        # text.bind("<Double-1>", self.text_double_click)

        var = ttk.Scrollbar(xfrm, orient=tk.HORIZONTAL)
        var.pack(side=tk.BOTTOM, fill=tk.X)
        var.config(command=text.xview)
        text.configure(xscrollcommand=var.set)

        # right-click menu
        m = tk.Menu(frame, tearoff=0)
        m.add_command(label="edit Text", command=self.edit_expression)
        m.add_command(label="view Metadata", command=self.view_metadata)

        def menu_popup(event):
            post_right_click_menu(m, event.x_root, event.y_root)
        text.bind("<Button-3>", menu_popup)
        return text

    def ini_terminal(self, frame: tk.Misc):
        # Terminal
        frm = ttk.Frame(frame)
        frm.pack(side=tk.BOTTOM, fill=tk.X, expand=tk.YES)

        tfrm = ttk.Frame(frm, relief=tk.RIDGE)
        tfrm.pack(side=tk.TOP, fill=tk.BOTH)

        text_chars, text_lines = get_text_size(self.root, self.config)
        terminal = tk.Text(tfrm, state=tk.DISABLED, wrap=tk.NONE, height=3, width=text_chars)
        terminal.pack(side=tk.LEFT, fill=tk.X, expand=tk.NO)

        var = ttk.Scrollbar(tfrm, orient=tk.VERTICAL)
        var.pack(side=tk.LEFT, fill=tk.Y)
        var.config(command=terminal.yview)
        terminal.configure(yscrollcommand=var.set)

        efrm = ttk.Frame(frm, relief=tk.GROOVE)
        efrm.pack(side=tk.TOP, fill=tk.BOTH)

        var = ttk.Label(efrm, text='>>')
        var.pack(side=tk.LEFT)
        var = ttk.Entry(efrm, textvariable=self.terminal_entry)
        var.pack(side=tk.LEFT, expand=tk.YES, fill=tk.BOTH)
        var.bind('<Return>', self.fun_terminal)
        var.bind('<KP_Enter>', self.fun_terminal)
        var.bind('<Up>', self.fun_terminal_history_up)
        var.bind('<Down>', self.fun_terminal_history_down)

        var = ttk.Button(efrm, text='CLS', command=self.fun_terminal_cls)
        var.pack(side=tk.LEFT)
        return terminal

    def ini_notebooks(self, frame: tk.Misc):
        frm = ttk.Frame(frame)
        frm.pack(side=tk.BOTTOM, fill=tk.X, expand=tk.YES, padx=6)

        menu = ttk.OptionMenu(frm, self.notebook)
        menu.pack(side=tk.LEFT)
        ttk.Button(frm, text='Open', command=self.run_notebook).pack(side=tk.LEFT)
        ttk.Button(frm, text='Reprocess', command=self.reprocess_notebook).pack(side=tk.LEFT)
        return menu

    def update_data_from_file(self, filename: str, hdf_map: hdfmap.NexusMap | None = None):
        self.filename = filename
        self.map = create_nexus_map(self.filename) if hdf_map is None else hdf_map
        self.update_text()
        self.notebooks = {
            os.path.basename(file): file
            for file in get_scan_notebooks(filename)
        }
        if self.notebooks:
            first_notebook = next(iter(self.notebooks))
            self.combo_notebook.set_menu(first_notebook, *self.notebooks)
        else:
            self.combo_notebook.set_menu('None')

    def update_text(self):
        try:
            with hdfmap.load_hdf(self.filename) as hdf:
                txt = self.map.format_hdf(hdf, self._text_expression)
            self.textbox.configure(state=tk.NORMAL)
            self.textbox.delete('1.0', tk.END)
            self.textbox.insert('1.0', txt)
            self.textbox.configure(state=tk.DISABLED)
        except Exception as e:
            show_error(f"Error:\n{e}", parent=self.root, raise_exception=False)

    def edit_expression(self):
        """Double-click on text display => open config str"""
        self._text_expression = EditText(self._text_expression, self.root).show()
        self.update_text()

    def view_metadata(self):
        from ..apps.namespace_select import create_metadata_selector
        hdf_map = self.map or create_nexus_map(self.filename)
        create_metadata_selector(hdf_map, self.root, self.config)

    def run_notebook(self):
        from mmg_toolbox.utils.nb_runner import view_jupyter_notebook, view_notebook_html
        notebook = self.notebook.get()
        if notebook not in self.notebooks:
            return
        filename = self.notebooks[notebook]
        html = filename.replace('.ipynb', '.html')
        if os.path.isfile(html):
            view_jupyter_notebook(html)
        else:
            # generate html in TMP
            view_notebook_html(filename)

    def reprocess_notebook(self):
        """Copy notebook to processing folder"""
        from mmg_toolbox.utils.nb_runner import reprocess_notebook
        notebook = self.notebook.get()
        if notebook not in self.notebooks:
            return
        filename = self.notebooks[notebook]

        response = askyesnocancel(
            title="NeXus Data Viewer",
            message=f"This will copy {notebook} and start a jupyter server\nYes will copy to processing, No to TMP",
            parent=self.root,
        )
        if response:
            print('Yes')
            reprocess_notebook(filename)
        elif response is None:
            print('Cancel')
            return
        else:
            print('No')
            reprocess_notebook(filename, output_folder=TMPDIR)

    def set_terminal(self):
        self.terminal_entry.set(self.terminal_history[self.terminal_history_index])

    def fun_terminal(self, event=None):
        if self.filename is None:
            return
        expression = self.terminal_entry.get()
        out_str = f"\n>>> {expression}\n"
        try:
            # TODO: replace with asteval.Interpreter (maybe in hdfmap V1.1)
            with hdfmap.load_hdf(self.filename) as hdf:
                out = self.map.eval(hdf, expression)
            self.terminal_history.insert(1, expression)
            self.terminal_history_index = 0
            self.set_terminal()
        except NameError as ne:
            out = ne
        out_str += f"{out}\n"
        self.terminal.configure(state=tk.NORMAL)
        self.terminal.insert(tk.END, out_str)
        self.terminal.see(tk.END)
        self.terminal.configure(state=tk.DISABLED)

    def fun_terminal_cls(self, event=None):
        # print('deleting')
        self.terminal.configure(state=tk.NORMAL)
        self.terminal.delete('1.0', tk.END)
        self.terminal.configure(state=tk.DISABLED)

    def fun_terminal_history_up(self, event=None):
        if len(self.terminal_history) > self.terminal_history_index:
            self.terminal_history_index += 1
            self.set_terminal()

    def fun_terminal_history_down(self, event=None):
        self.terminal_history_index -= 1
        self.set_terminal()
